import bpy
import numpy as np
from mathutils import Vector, Euler
import bmesh
def cart2sph(x, y, z):
    hxy = np.hypot(x, y)
    r = np.hypot(hxy, z)
    el = np.arctan2(z, hxy)
    az = np.arctan2(y, x)
    return az, el, r
def azel_fromRotMatrix_dir(Mat,Dir):
    inv = Mat.inverted()
    A=Euler(np.radians((90,0,-90)), 'XYZ').to_matrix()
    dir = inv @ Dir
    dir = A @ dir
    az, el, r = cart2sph(dir.x, dir.y, dir.z)
    return az , el

class BlenderGeometry:
    def get_Target_Position(self, scene, depsgraph):
        face_center_all = []
        HashFaceIndex_ScattersGeo={}
        for obj in scene.objects:
            if obj.type == 'MESH':
                if obj.name.startswith('Probe_')==False:
                    if "RCS0" not in obj:
                        obj["RCS0"] = 1.0
                    RCS0 = obj["RCS0"]
                    if "Backscatter N" not in obj:
                        obj["Backscatter N"] = 1
                    Backscatter_N = obj["Backscatter N"]
                    if "Backscatter Dev (deg)" not in obj:
                        obj["Backscatter Dev (deg)"] = 0.0
                    Backscatter_dev = obj["Backscatter Dev (deg)"]
                    bm = bmesh.new()
                    bm.from_object(obj, depsgraph)
                    bm.transform(obj.matrix_world)
                    obj_hash = hash(obj)
                    HashFaceIndex_ScattersGeo[obj_hash]={}
                    for face in bm.faces:
                        face_center = face.calc_center_median()
                        fn = obj.matrix_world.to_3x3() @ face.normal
                        fn.normalize()
                        face_center_all.append([face_center,obj_hash,face.index,face.calc_area(),fn,RCS0,Backscatter_N,Backscatter_dev])
                        # print(face_center,face.normal,fn)
                        HashFaceIndex_ScattersGeo[obj_hash][face.index]=len(face_center_all)-1
                    bm.free()
        return face_center_all,HashFaceIndex_ScattersGeo
    def get_Suite_Position(self, suite_information,depsgraph):
        Suite_Position = []
        for suite in suite_information:  # suite_planes[i]['radars'][j]["TXs"]

            suite_info={}

            suite_info['Radar'] = []

            for radar in suite['Radar']:

                txPos = []
                txDir = []
                rxPos = []
                rxDir = []
                txpower = []
                TX_PatternType=[]
                TX_PatternMaxGain=[]
                TX_PatternAzimuthBeamWidth = []
                TX_PatternElevationBeamWidth=[]
                RX_PatternType=[]
                RX_PatternMaxGain=[]
                RX_PatternAzimuthBeamWidth = []
                RX_PatternElevationBeamWidth=[]
                LightSpeed=299792458.0
                WaveLength=LightSpeed/radar['GeneralRadarSpec_Object']["Center_Frequency_GHz"]/1e9
                Pt = 10**((radar['GeneralRadarSpec_Object']['Transmit_Power_dBm']-30)/10)
                Pt /= len(radar['TX'])
                Gt = 10**((radar['GeneralRadarSpec_Object']['Transmit_Antenna_Element_Gain_db'])/10)
                Gr = 10**((radar['GeneralRadarSpec_Object']['Receive_Antenna_Element_Gain_db'])/10)

                for c in radar['TX']:
                    global_location, global_rotation, global_scale = c.matrix_world.decompose()
                    txPos.append(global_location)
                    txDir.append(global_rotation)
                    txpower.append(Pt)
                    TX_PatternType.append(radar['GeneralRadarSpec_Object']["Transmit_Antenna_Element_Pattern"])
                    # TX_PatternType.append("Omni")
                    TX_PatternMaxGain.append(Gt)
                    TX_PatternAzimuthBeamWidth.append(radar['GeneralRadarSpec_Object']["Transmit_Antenna_Element_Azimuth_BeamWidth_deg"])
                    TX_PatternElevationBeamWidth.append(radar['GeneralRadarSpec_Object']["Transmit_Antenna_Element_Elevation_BeamWidth_deg"])
                for c in radar['RX']:
                    global_location, global_rotation, global_scale = c.matrix_world.decompose()
                    rxPos.append(global_location)
                    rxDir.append(global_rotation)
                    # RX_PatternType.append("Omni")
                    RX_PatternType.append(radar['GeneralRadarSpec_Object']["Receive_Antenna_Element_Pattern"])
                    RX_PatternMaxGain.append(Gr)
                    RX_PatternAzimuthBeamWidth.append(radar['GeneralRadarSpec_Object']["Receive_Antenna_Element_Azimuth_BeamWidth_deg"])
                    RX_PatternElevationBeamWidth.append(radar['GeneralRadarSpec_Object']["Receive_Antenna_Element_Elevation_BeamWidth_deg"])


                suite_info['Radar'].append({'TX-Position':txPos,'RX-Position':rxPos,
                                            'TX-Velocity':[],'RX-Velocity':[],
                                            'TX-Direction':txDir,'RX-Direction':rxDir,
                                            'TX-Direction_Next':[],'RX-Direction_Next':[],
                                            'TX-Power':txpower,'TX-PatternType':TX_PatternType,'TX-PatternMaxGain':TX_PatternMaxGain,
                                            'TX-PatternAzimuthBeamWidth':TX_PatternAzimuthBeamWidth,'TX-PatternElevationBeamWidth':TX_PatternElevationBeamWidth,
                                            'RX-PatternType':RX_PatternType,'RX-PatternMaxGain':RX_PatternMaxGain,
                                            'RX-PatternAzimuthBeamWidth':RX_PatternAzimuthBeamWidth,'RX-PatternElevationBeamWidth':RX_PatternElevationBeamWidth,
                                            'WaveLength':WaveLength
                                            })

                    # az,el = azel_fromRotMatrix_dir(TX-Direction,Exact_start-EndRay0)
            suite_info['RIS'] = []
            LightSpeed = 299792458.0
            for ris in suite['RIS']:
                Pos = []
                Dir = []
                PhaseAmp = []
                PatternType=[]
                PatternMaxGain=[]
                PatternAzimuthBeamWidth=[]
                PatternElevationBeamWidth=[]
                WaveLength=LightSpeed/70e9
                for c in ris:
                    global_location, global_rotation, global_scale = c.matrix_world.decompose()
                    Pos.append(global_location)
                    Dir.append(global_rotation)
                    PhaseAmp.append(np.array([c['amplitude'],c['phase']]))
                    PatternType.append("Omni")
                    PatternMaxGain.append(1)
                    PatternAzimuthBeamWidth.append(60)
                    PatternElevationBeamWidth.append(80)

                suite_info['RIS'].append({'Position':Pos,'Velocity':[],'Direction':Dir,'PhaseAmplitude':PhaseAmp,
                                          'PatternType':PatternType,'PatternMaxGain':PatternMaxGain,
                                          'PatternAzimuthBeamWidth':PatternAzimuthBeamWidth,'PatternElevationBeamWidth':PatternElevationBeamWidth,
                                          'WaveLength':WaveLength})

            suite_info['Probe'] = []

            # print(suite['Probe'])
            for probe in suite['Probe']:
                Pos = []
                for obj in probe:
                    # print(obj)
                    # obj.hide_viewport = False
                    # probe_points = []
                    # if obj.type == 'MESH':
                    #     object_id = hash(obj.name)
                    #     bm = bmesh.new()
                    #     bm.verts.ensure_lookup_table()
                    #     bm.from_object(obj, depsgraph)
                    #     bm.transform(obj.matrix_world)
                    #     mesh = bpy.data.meshes.new('new_mesh')
                    #     bm.to_mesh(mesh)
                    #     for face in bm.faces:
                    #         # probe_points.append(face.calc_center_median())
                    #         probe_points.append([face.calc_center_median(),[vert.co.copy() for vert in face.verts],0])
                    #     bm.free()
                    # obj.hide_viewport = True
                    global_location, global_rotation, global_scale = obj.matrix_world.decompose()
                    
                    Pos.append(global_location)

                suite_info['Probe'].append(Pos)

            Suite_Position.append(suite_info)
        return Suite_Position

    def get_Position_Velocity(self, scene, suite_information, frame, frame_jump_for_velocity):

        # for isuite,suiteobject in enumerate(suite_information):
        #   for probes in suiteobject['Probe']:
        #     for probe in probes:
        #       probe.hide_viewport = False

        JumpTime = frame_jump_for_velocity / float(scene.render.fps)
        frame_jump_for_velocity = int(frame_jump_for_velocity)
        scene.frame_set(frame+frame_jump_for_velocity)
        bpy.context.view_layer.update()
        depsgraph = bpy.context.evaluated_depsgraph_get()
        face_center_all_next,_ = self.get_Target_Position( scene, depsgraph)
        Suite_Position_next = self.get_Suite_Position(suite_information,depsgraph)

        scene.frame_set(frame)
        bpy.context.view_layer.update()
        depsgraph = bpy.context.evaluated_depsgraph_get()
        face_center_all,HashFaceIndex_ScattersGeo = self.get_Target_Position( scene, depsgraph)
        Suite_Position = self.get_Suite_Position(suite_information,depsgraph)

        for i,s in enumerate(Suite_Position):
            for j,r in enumerate(s['Radar']):
                Suite_Position[i]['Radar'][j]['TX-Velocity']=[]
                for k,t in enumerate(r['TX-Position']):
                    v = Suite_Position_next[i]['Radar'][j]['TX-Position'][k] - t
                    Suite_Position[i]['Radar'][j]['TX-Velocity'].append(v/JumpTime)
                Suite_Position[i]['Radar'][j]['TX-Direction_Next']=Suite_Position_next[i]['Radar'][j]['TX-Direction']
                Suite_Position[i]['Radar'][j]['RX-Velocity']=[]
                for k,t in enumerate(r['RX-Position']):
                    v = Suite_Position_next[i]['Radar'][j]['RX-Position'][k] - t
                    Suite_Position[i]['Radar'][j]['RX-Velocity'].append(v/JumpTime)
                Suite_Position[i]['Radar'][j]['RX-Direction_Next']=Suite_Position_next[i]['Radar'][j]['RX-Direction']

            for iris , ris in enumerate(Suite_Position[i]['RIS']):
              # Suite_Position[isuite]['RIS'][iris]['Position'][iriselement]
                Suite_Position[i]['RIS'][iris]['Velocity']=[]
                for iriselement ,ris_element_position in enumerate(ris['Position']):
                    v = Suite_Position_next[i]['RIS'][iris]['Position'][iriselement] - ris_element_position
                    Suite_Position[i]['RIS'][iris]['Velocity'].append(v/JumpTime)
        v = []
        for i,c in enumerate(face_center_all):
            v.append((face_center_all_next[i][0] - c[0]) / JumpTime)

        # for isuite,suiteobject in enumerate(suite_information):
        #   for probes in suiteobject['Probe']:
        #     for probe in probes:
        #       probe.hide_viewport = True

        return Suite_Position,face_center_all,HashFaceIndex_ScattersGeo,v





import sensingsp as ssp
import numpy as np
from mathutils import Vector
# import matplotlib
# matplotlib.use('Qt5Agg')
from matplotlib import pyplot as plt
import cv2
import os
import bpy

def make_simple_scenario():
    predefine_movingcube_6843()
def run_simple_chain():
    processing_1()
def add_scenario(st):
    if st=='2 Cubes + 6843':
        predefine_movingcube_6843()
    if st == 'Pattern SISO':
        ssp.utils.delete_all_objects()
        ssp.utils.define_settings()
        R=10
        NF = 360*2
        cube = ssp.environment.add_cube(location=Vector((R, 0, 0)), direction=Vector((1, 0, 0)), scale=Vector((.5, .5, .5)), subdivision=0)
        cube["RCS0"]=1
        bpy.ops.object.empty_add(type='PLAIN_AXES', align='WORLD', location=Vector((0, 0, 0)), rotation=Vector((0, 0, 0)), scale=(.01, .01, .01))
        empty = bpy.context.object
        cube.parent = empty
        empty.rotation_euler  = (0,0,2*np.pi)
        empty.keyframe_insert(data_path="rotation_euler", frame=1)
        empty.rotation_euler  = (0,0,0)
        empty.keyframe_insert(data_path="rotation_euler", frame=NF/2)
        empty.rotation_euler  = (0,2*np.pi,0)
        empty.keyframe_insert(data_path="rotation_euler", frame=NF)
        for fcurve in empty.animation_data.action.fcurves:
            for keyframe in fcurve.keyframe_points:
                keyframe.interpolation = 'LINEAR'
        ssp.integratedSensorSuite.define_suite(0, location=Vector((0, 0, 0)), rotation=Vector((0, 0, 0)))
        radar = ssp.radar.utils.predefined_array_configs_SISO(isuite=0, iradar=0, location=Vector((0, 0,0)), rotation=Vector((np.pi/2,0, -np.pi/2)), f0=70e9)
        ssp.utils.set_frame_start_end(start=1,end=NF)
        
    if st=='Hand Gesture + 1642':
        ssp.utils.delete_all_objects()
        # try:
        #     ssp.utils.open_Blend(os.path.join(ssp.config.temp_folder, 'arm_to_left 4.blend'))
        # except Exception as X:
        #     1
        blend_file_path = os.path.join(ssp.config.temp_folder, 'arm_to_left 4.blend')

        # The directory inside the .blend file to access objects
        object_directory = blend_file_path + "/Object/"

        # Append all objects from the file
        with bpy.data.libraries.load(blend_file_path) as (data_from, data_to):
            # List all objects available in the .blend file
            print("Available objects:", data_from.objects)
            # Select objects you want to append (replace with specific names if needed)
            data_to.objects = data_from.objects  # Load all objects

        # Link the loaded objects to the current scene
        for obj in data_to.objects:
            if obj is not None:  # Avoid null references
                bpy.context.scene.collection.objects.link(obj)
        ssp.utils.define_settings()
        ssp.integratedSensorSuite.define_suite(0, location=Vector((0, 0, 0)), rotation=Vector((0, 0, 0)))
        radar = ssp.radar.utils.predefined_array_configs_TI_AWR1642(isuite=0, iradar=0, location=Vector((0, 0,0)), rotation=Vector((np.pi/2,0, -np.pi/2)), f0=76e9)
        rangeResolution,radialVelocityResolution,N_ADC,ChirpTimeMax,CentralFreq=0.039,0.13,256,60e-6,76e9

        ssp.radar.utils.FMCW_Chirp_Parameters(rangeResolution,N_ADC,ChirpTimeMax,radialVelocityResolution,CentralFreq)

        ssp.radar.utils.set_FMCW_Chirp_Parameters(radar,slope=64.06,fsps=4.2,N_ADC=256,NPulse=256,PRI_us=60)
        radar['RangeFFT_OverNextP2'] = 0
        radar['Range_End']=100*.6/9.98 
        radar['DopplerFFT_OverNextP2']=0
        radar['CFAR_RD_alpha']=15
        radar['CFAR_Angle_alpha']=2
        radar['Transmit_Antenna_Element_Pattern']='Directional-Sinc' # add rect pattern
        radar["Transmit_Antenna_Element_Azimuth_BeamWidth_deg"] = 60
        radar["Transmit_Antenna_Element_Elevation_BeamWidth_deg"] = 60

        ssp.utils.set_frame_start_end(start=1,end=30)
        ssp.utils.save_Blender()

    if st == "Hand Gesture + 3 Xethru Nature paper":
        ssp.utils.delete_all_objects()
        ssp.utils.define_settings()
        ssp.integratedSensorSuite.define_suite(0, location=Vector((0, 0, 0)), rotation=Vector((0, 0, 0)))
        radar1 = ssp.radar.utils.predefined_array_configs_SISO(isuite=0, iradar=0, location=Vector((0, 0,0)), rotation=Vector((np.pi/2,0, -np.pi/2)), f0=76e9)
        radar2 = ssp.radar.utils.predefined_array_configs_SISO(isuite=0, iradar=1, location=Vector((1, 0,0)), rotation=Vector((np.pi/2,0, np.pi/2)), f0=76e9)
        radar3 = ssp.radar.utils.predefined_array_configs_SISO(isuite=0, iradar=2, location=Vector((.5, 0,.5)), rotation=Vector((0,0, 0)), f0=76e9)
        # During data acquisition, examiners recorded a subject repeating a particular hand gesture for 450 seconds, corresponding to 9000 (slow-time) rows. There is 1 complete gesture motion in 90 slow-time frames. As such, each radar signal matrix contains 100 complete hand gesture motion samples. The range of each UWB radar is 1.2 meters, corresponding to 189 fast-time bins.
        gesture_motion_slowTimeFrames = 90
        recordedTimePerSample = 4.5
        radarRange = 1.2
        fast_time_bins = 189
        rangebins = radarRange / fast_time_bins
        for radar in [radar1,radar2,radar3]:
            radar['RadarMode']='Pulse'
            radar['PulseWaveform']='UWB'
            radar['N_ADC']  = fast_time_bins
            radar['Fs_MHz'] =  ssp.LightSpeed / (2*rangebins) / 1e6
            radar['PRI_us']= 1e6*recordedTimePerSample/gesture_motion_slowTimeFrames
            radar['NPulse'] = gesture_motion_slowTimeFrames
            radar['Range_End']=100
        FramesNumber=int(1.1*recordedTimePerSample*bpy.context.scene.render.fps)
        blend_file_path = os.path.join(ssp.config.temp_folder, 'RealisticHand.blend')
        object_directory = blend_file_path + "/Object/"
        with bpy.data.libraries.load(blend_file_path) as (data_from, data_to):
            data_to.objects = data_from.objects  # Load all objects

        for obj in data_to.objects:
            if obj is not None:  # Avoid null references
                bpy.context.scene.collection.objects.link(obj)
        # ssp.utils.decimate_scene_all(.1)    
        ssp.utils.set_RayTracing_balanced()
        ssp.utils.set_frame_start_end(start=1,end=FramesNumber)
        ssp.utils.save_Blender()

    if st == 'Ray Tracing 1':
        ssp.utils.delete_all_objects()
        ssp.utils.define_settings()
        cube = ssp.environment.add_cube(location=Vector((1.5, -.5, 0)), direction=Vector((1, 0, 0)), scale=Vector((.5, .5, .05)), subdivision=0)
        cube = ssp.environment.add_cube(location=Vector((2.5, 1.5, 0)), direction=Vector((1, 0, 0)), scale=Vector((.5, .5, .05)), subdivision=0)
        cube = ssp.environment.add_cube(location=Vector((4.5, 0.5, 0)), direction=Vector((1, 0, 0)), scale=Vector((.5, 1.5, .05)), subdivision=0)
        cube = ssp.environment.add_cube(location=Vector((2.5, -1.5, 0)), direction=Vector((0, 1, 0)), scale=Vector((.5, .5, .05)), subdivision=0)
        
        for obj in bpy.context.scene.objects:
            if obj.type == 'MESH':
                if obj.name.startswith('Probe_')==False:
                    if "RCS0" not in obj:
                        obj["RCS0"] = 1.0
                    RCS0 = obj["RCS0"]
                    if "Backscatter N" not in obj:
                        obj["Backscatter N"] = 1
                    Backscatter_N = obj["Backscatter N"]
                    if "Backscatter Dev (deg)" not in obj:
                        obj["Backscatter Dev (deg)"] = 0.0
        
        
        ssp.integratedSensorSuite.define_suite(0, location=Vector((0, 0, 0)), rotation=Vector((0, 0, 0)))
        radar = ssp.radar.utils.predefined_array_configs_SISO(isuite=0, iradar=0, location=Vector((0, 0,0)), rotation=Vector((np.pi/2,0, -np.pi/2)), f0=70e9)
        
        obj = bpy.data.objects.get("RX_0_0_1_0_00001")
        if obj:
            obj.location = (1, 0, -3)  # Replace x, y, z with the desired coordinates
            obj.scale = (1, 1, 1)
    if st == 'Ray Tracing 2':
        ssp.utils.delete_all_objects()
        ssp.utils.define_settings()
        cube = ssp.environment.add_cube(location=Vector((1.5, -.5, 0)), direction=Vector((1, 0, 0)), scale=Vector((.5, .5, .05)), subdivision=0)
        cube = ssp.environment.add_cube(location=Vector((2.5, 1.5, 0)), direction=Vector((1, 0, 0)), scale=Vector((.5, .5, .05)), subdivision=0)
        cube = ssp.environment.add_cube(location=Vector((4.5, 0.5, 0)), direction=Vector((1, 0, 0)), scale=Vector((.5, 1.5, .05)), subdivision=0)
        cube = ssp.environment.add_cube(location=Vector((2.5, -1.5, 0)), direction=Vector((0, 1, 0)), scale=Vector((.5, .5, .05)), subdivision=0)
        
        for obj in bpy.context.scene.objects:
            if obj.type == 'MESH':
                if obj.name.startswith('Probe_')==False:
                    if "RCS0" not in obj:
                        obj["RCS0"] = 1.0
                    RCS0 = obj["RCS0"]
                    if "Backscatter N" not in obj:
                        obj["Backscatter N"] = 1
                    Backscatter_N = obj["Backscatter N"]
                    if "Backscatter Dev (deg)" not in obj:
                        obj["Backscatter Dev (deg)"] = 0.0
        
        
        ssp.integratedSensorSuite.define_suite(0, location=Vector((0, 0, 0)), rotation=Vector((0, 0, 0)))
        radar = ssp.radar.utils.predefined_array_configs_SISO(isuite=0, iradar=0, location=Vector((0, 0,0)), rotation=Vector((np.pi/2,0, -np.pi/2)), f0=70e9)
    if st == 'Ray Tracing 3':
        ssp.utils.delete_all_objects()
        ssp.utils.define_settings()
        cube = ssp.environment.add_cube(location=Vector((1.5, -.5, 0)), direction=Vector((1, 0, 0)), scale=Vector((.5, .5, .05)), subdivision=0)
        cube = ssp.environment.add_cube(location=Vector((2.5, 1.5, 0)), direction=Vector((1, 0, 0)), scale=Vector((.5, .5, .05)), subdivision=0)
        cube = ssp.environment.add_cube(location=Vector((4.5, 0.5, 0)), direction=Vector((1, 0, 0)), scale=Vector((.5, 1.5, .05)), subdivision=0)
        cube = ssp.environment.add_cube(location=Vector((2.5, -1.5, 0)), direction=Vector((0, 1, 0)), scale=Vector((.5, .5, .05)), subdivision=0)
        
        for obj in bpy.context.scene.objects:
            if obj.type == 'MESH':
                if obj.name.startswith('Probe_')==False:
                    if "RCS0" not in obj:
                        obj["RCS0"] = 1.0
                    RCS0 = obj["RCS0"]
                    if "Backscatter N" not in obj:
                        obj["Backscatter N"] = 1
                    Backscatter_N = obj["Backscatter N"]
                    if "Backscatter Dev (deg)" not in obj:
                        obj["Backscatter Dev (deg)"] = 0.0
        
        
        ssp.integratedSensorSuite.define_suite(0, location=Vector((0, 0, 0)), rotation=Vector((0, 0, 0)))
        radar = ssp.radar.utils.predefined_array_configs_TI_AWR1642(isuite=0, iradar=0, location=Vector((0, 0,0)), rotation=Vector((np.pi/2,0, -np.pi/2)), f0=70e9)
        
def sim_scenario(st):
    if st=='2 Cubes + 6843':
        processing_1()
    if st=='Hand Gesture + 1642':
        processing_HandGesture1642()
    if st == 'Pattern SISO':
        ssp.utils.trimUserInputs() 
        ssp.config.restart()
        ssp.config.setDopplerProcessingMethod_FFT_Winv(0)
        # # plt.ion() 
        # fig, FigsAxes = plt.subplots(2,3)
        # FigsAxes[1, 2] = fig.add_subplot(2, 3, 6, projection='3d')
        
        pat=[]
        while ssp.config.run():
            path_d_drate_amp = ssp.raytracing.Path_RayTracing_frame()
            alld = []
            for itx in range(len(path_d_drate_amp[0][0][0][0][0])):
                for irx in range(len(path_d_drate_amp[0][0])):
                    for d_drate_amp in path_d_drate_amp[0][0][irx][0][0][itx]:
                        alld.append(d_drate_amp[0:3])
            alld = np.array(alld)
            if len(alld)>0:
                pat.append([ssp.config.CurrentFrame,alld[0][2]])
            else:
                pat.append(0)
            ssp.utils.increaseCurrentFrame()
        f=np.array([x[0] for x in pat])
        p=np.array([x[1] for x in pat])
        theta = np.arange(360)
        p/=p.max()
        plt.figure()
        plt.plot(theta,p[:360])
        plt.plot(theta[:len(p[360:])],p[360:])
        plt.figure()
        plt.plot(theta,20*np.log10(p[:360]))
        plt.plot(theta[:len(p[360:])],20*np.log10(p[360:]))
        plt.show()
    
    if st == 'Ray Tracing 1' or st == 'Ray Tracing 2' or st == 'Ray Tracing 3' :
        ssp.environment.scenarios.raytracing_test()    
    
    if st == "Hand Gesture + 3 Xethru Nature paper":
        ssp.utils.trimUserInputs()  
        ssp.config.restart()
        # ssp.config.DopplerProcessingMethod_FFT_Winv = (True)

        # ssp.config.directReceivefromTX =  False 
        # ssp.config.RadarRX_only_fromscatters_itsTX = True
        # ssp.config.RadarRX_only_fromitsTX = True
        # ssp.config.Radar_TX_RX_isolation = True

        ssp.utils.useCUDA(False)

        while ssp.config.run():
            path_d_drate_amp = ssp.raytracing.Path_RayTracing_frame()
            Signals = ssp.integratedSensorSuite.SensorsSignalGeneration_frame(path_d_drate_amp)
            for isuite, radarSpecifications in enumerate(ssp.RadarSpecifications):
                for iradar, specifications in enumerate(radarSpecifications):
                    for XRadar, timeX in Signals[isuite]['radars'][iradar]:
                        # Range and Doppler processing
                        plt.figure()
                        plt.imshow(np.abs(XRadar[:,:,0]), aspect='auto', cmap='viridis')
                        plt.title(f'{iradar}')
            print(f'Processed frame = {ssp.config.CurrentFrame}')
            ssp.utils.increaseCurrentFrame()
        plt.show()
    
def predefine_movingcube_6843(interpolation_type='LINEAR'):
    #     Blender provides several interpolation types, and as of recent versions, there are 6 main interpolation options:

    # CONSTANT ('CONSTANT')
    # The value remains constant between keyframes, resulting in a stepped change at each keyframe.
    # LINEAR ('LINEAR')
    # The value changes linearly between keyframes, resulting in a straight transition between the keyframe points.
    # BEZIER ('BEZIER')
    # The value follows a bezier curve between keyframes, allowing for smooth transitions with adjustable handles for fine control over the curve.
    # SINE ('SINE')
    # The value follows a sine curve between keyframes, creating smooth, wave-like transitions.
    # QUAD ('QUAD')
    # The value follows a quadratic curve, which can create an accelerating or decelerating effect.
    # CUBIC ('CUBIC')
    # The value follows a cubic curve between keyframes, providing a slightly more complex smooth transition than quadratic.
    ssp.utils.delete_all_objects()
    ssp.utils.define_settings()
    cubex = ssp.environment.add_cube(location=Vector((2.7, np.sqrt(3**2-2.7**2), 0)), direction=Vector((1,0, 0)), scale=Vector((.1, .1, .1)), subdivision=0)
    cubex["RCS0"]=1
    cube = ssp.environment.add_cube(location=Vector((0, 0, 0)), direction=Vector((1, 0, 0)), scale=Vector((.1, .1, .1)), subdivision=0)
    cube["RCS0"]=1
    cube.location = (3,0,0)
    cube.keyframe_insert(data_path="location", frame=1)
    cube.location = (3, 3,0)
    cube.keyframe_insert(data_path="location", frame=30)
    cube.location = (3, -3,0)
    cube.keyframe_insert(data_path="location", frame=100)
    for fcurve in cube.animation_data.action.fcurves:
        for keyframe in fcurve.keyframe_points:
            keyframe.interpolation = interpolation_type
    ssp.integratedSensorSuite.define_suite(0, location=Vector((0, 0, 0)), rotation=Vector((0, 0, 0)))
    
    
    radar = ssp.radar.utils.predefined_array_configs_TI_IWR6843(isuite=0, iradar=0, location=Vector((0, 0,0)), rotation=Vector((np.pi/2,0, -np.pi/2)), f0=70e9)
    # radar = ssp.radar.utils.predefined_array_configs_TI_IWR6843_az(isuite=0, iradar=0, location=Vector((0, 0,0)), rotation=Vector((np.pi/2,0, -np.pi/2)), f0=70e9)
    # radar = ssp.radar.utils.predefined_array_configs_LinearArray(isuite=0, iradar=0, location=Vector((0, 0,0)), rotation=Vector((np.pi/2,0, -np.pi/2)), f0=70e9,LinearArray_TXPos=[0],LinearArray_RXPos=[i*3e8/70e9/2 for i in range(30)])
    
    
    
    # radar['RF_AnalogNoiseFilter_Bandwidth_MHz']=0
    
    ssp.utils.set_frame_start_end(start=1,end=2)
    ssp.utils.useCUDA()
    # print(f"rangeResolution maxUnambigiousRange = {ssp.radar.utils.rangeResolution_and_maxUnambigiousRange(radar)}")

def predefine_TruckHuman_SameRG():
    ssp.utils.delete_all_objects()
    if 0:
        ssp.utils.open_Blend(os.path.join(ssp.config.temp_folder, "TruckHuman.blend"))
    # ssp.utils.open_Blend(os.path.join(ssp.config.temp_folder, "SimpleCube_y05.blend"))
    else:
        cube = ssp.environment.add_cube(location=Vector((20, 0, 0)), direction=Vector((1, 0, 0)), scale=Vector((.1, .1, .1)), subdivision=0)
        cube["RCS0"]=1e5
        t=np.deg2rad(2)
        cube = ssp.environment.add_cube(location=Vector((20*np.cos(t), 20*np.sin(t), 0)), direction=Vector((1, 0, 0)), scale=Vector((.1, .1, .1)), subdivision=0)
        cube["RCS0"]=1e5
        # cube.location = (20,0,0)
        # cube.keyframe_insert(data_path="location", frame=1)
        # cube.location = (20, 3,0)
        # cube.keyframe_insert(data_path="location", frame=50)
        # cube.location = (20, 0,0)
        # cube.keyframe_insert(data_path="location", frame=100)
        
    
    ssp.utils.define_settings()
    ssp.integratedSensorSuite.define_suite(0, location=Vector((0, 0, 0)), rotation=Vector((0, 0, 0)))
    radar = ssp.radar.utils.predefined_array_configs_DARWu(isuite=0, iradar=0, location=Vector((0, 0,0)), rotation=Vector((np.pi/2,0, -np.pi/2)), f0=70e9)
    # radar = ssp.radar.utils.predefined_array_configs_TI_IWR6843(isuite=0, iradar=0, location=Vector((0, 0,0)), rotation=Vector((np.pi/2,0, -np.pi/2)), f0=70e9)
    # radar = ssp.radar.utils.predefined_array_configs_LinearArray(isuite=0, iradar=0, location=Vector((0, 0,0)), rotation=Vector((np.pi/2,0, -np.pi/2)), 
    #                                                              f0=70e9
    #                                                              ,LinearArray_TXPos =[(i)*4*3e8/70e9/2 for i in range(1)],
    #                                                             LinearArray_RXPos =[(i+10)*3e8/70e9/2 for i in range(4)])
    # radar = ssp.radar.utils.predefined_array_configs_SISO(isuite=0, iradar=0, location=Vector((0, 0,0)), rotation=Vector((np.pi/2,0, -np.pi/2)), f0=70e9)
    
    Res,MaxR =    ssp.radar.utils.rangeResolution_and_maxUnambigiousRange(radar)
    radar['RF_AnalogNoiseFilter_Bandwidth_MHz']=2
    radar['DopplerProcessingMIMODemod']='Simple'
    radar['ADC_levels']=256*8
    radar['RangeFFT_OverNextP2']=1
    radar['DopplerFFT_OverNextP2']=0
    radar['AzFFT_OverNextP2']=5
    radar['NPulse']=6*64
    
# specifications['RangeFFT_OverNextP2'] = radarobject['GeneralRadarSpec_Object']['RangeFFT_OverNextP2']
# specifications['DopplerFFT_OverNextP2'] = radarobject['GeneralRadarSpec_Object']['']
# specifications['AzFFT_OverNextP2'] = radarobject['GeneralRadarSpec_Object']['AzFFT_OverNextP2']
# specifications['ElFFT_OverNextP2'] = radarobject['GeneralRadarSpec_Object']['ElFFT_OverNextP2']

    ssp.utils.set_frame_start_end(start=1,end=40)
    ssp.utils.useCUDA()
    ssp.utils.trimUserInputs() 
    ssp.config.restart()
    ssp.config.DopplerProcessingMethod_FFT_Winv = (True)

    ssp.config.directReceivefromTX =  False 
    ssp.config.RadarRX_only_fromscatters_itsTX = True
    ssp.config.RadarRX_only_fromitsTX = True
    ssp.config.Radar_TX_RX_isolation = True
    ssp.utils.save_Blender()
def predefine_TruckHuman_DiffRG():
    ssp.utils.delete_all_objects()
    if 1:
        ssp.utils.open_Blend(os.path.join(ssp.config.temp_folder, "TruckHumanModified.blend"))
    # ssp.utils.open_Blend(os.path.join(ssp.config.temp_folder, "SimpleCube_y05.blend"))
    else:
        cube = ssp.environment.add_cube(location=Vector((20, 20, 0)), direction=Vector((1, 0, 0)), scale=Vector((.1, .1, .1)), subdivision=0)
        cube["RCS0"]=1e6
        # cube.location = (20,0,0)
        # cube.keyframe_insert(data_path="location", frame=1)
        # cube.location = (20, 3,0)
        # cube.keyframe_insert(data_path="location", frame=50)
        # cube.location = (20, 0,0)
        # cube.keyframe_insert(data_path="location", frame=100)
        
    
    ssp.utils.define_settings()
    ssp.integratedSensorSuite.define_suite(0, location=Vector((0, 0, 0)), rotation=Vector((0, 0, 0)))
    radar = ssp.radar.utils.predefined_array_configs_DARWu(isuite=0, iradar=0, location=Vector((0, 0,0)), rotation=Vector((np.pi/2,0, -np.pi/2)), f0=70e9)
    # radar = ssp.radar.utils.predefined_array_configs_TI_IWR6843(isuite=0, iradar=0, location=Vector((0, 0,0)), rotation=Vector((np.pi/2,0, -np.pi/2)), f0=70e9)
    # radar = ssp.radar.utils.predefined_array_configs_LinearArray(isuite=0, iradar=0, location=Vector((0, 0,0)), rotation=Vector((np.pi/2,0, -np.pi/2)), 
    #                                                              f0=70e9
    #                                                              ,LinearArray_TXPos =[(i)*4*3e8/70e9/2 for i in range(1)],
    #                                                             LinearArray_RXPos =[(i+10)*3e8/70e9/2 for i in range(4)])
    # radar = ssp.radar.utils.predefined_array_configs_SISO(isuite=0, iradar=0, location=Vector((0, 0,0)), rotation=Vector((np.pi/2,0, -np.pi/2)), f0=70e9)
    
    Res,MaxR =    ssp.radar.utils.rangeResolution_and_maxUnambigiousRange(radar)
    radar['RF_AnalogNoiseFilter_Bandwidth_MHz']=20
    radar['DopplerProcessingMIMODemod']='Simple'
    radar['ADC_levels']=256*8
    radar['RangeFFT_OverNextP2']=1
    radar['DopplerFFT_OverNextP2']=0
    radar['AzFFT_OverNextP2']=5
    # radar['NPulse']=12
    
# specifications['RangeFFT_OverNextP2'] = radarobject['GeneralRadarSpec_Object']['RangeFFT_OverNextP2']
# specifications['DopplerFFT_OverNextP2'] = radarobject['GeneralRadarSpec_Object']['']
# specifications['AzFFT_OverNextP2'] = radarobject['GeneralRadarSpec_Object']['AzFFT_OverNextP2']
# specifications['ElFFT_OverNextP2'] = radarobject['GeneralRadarSpec_Object']['ElFFT_OverNextP2']

    ssp.utils.set_frame_start_end(start=1,end=40)
    ssp.utils.useCUDA()
    ssp.utils.trimUserInputs() 
    ssp.config.restart()
    ssp.config.DopplerProcessingMethod_FFT_Winv = (True)

    ssp.config.directReceivefromTX =  False 
    ssp.config.RadarRX_only_fromscatters_itsTX = True
    ssp.config.RadarRX_only_fromitsTX = True
    ssp.config.Radar_TX_RX_isolation = True
    ssp.utils.save_Blender()


def processing_1():
    ssp.utils.trimUserInputs() 
    ssp.config.restart()
    ssp.config.setDopplerProcessingMethod_FFT_Winv(0)
    # plt.ion() 
    fig, FigsAxes = plt.subplots(2,3)
    FigsAxes[1, 2] = fig.add_subplot(2, 3, 6, projection='3d')
    while ssp.config.run():
        path_d_drate_amp = ssp.raytracing.Path_RayTracing_frame()
        alld = []
        m = path_d_drate_amp[0][0][0][0][0][0][0][3]
        for itx in range(len(path_d_drate_amp[0][0][0][0][0])):
            for irx in range(len(path_d_drate_amp[0][0])):
                for d_drate_amp in path_d_drate_amp[0][0][irx][0][0][itx]:
                    # if d_drate_amp[3]==m:
                    alld.append(d_drate_amp[0:3])
        alld = np.array(alld)
        fig.suptitle(f'Frame: {ssp.config.CurrentFrame}')
        FigsAxes[0,2].cla()
        # FigsAxes[1,2].cla()
        # FigsAxes[0,2].plot(alld[:,0],'.')
        # FigsAxes[1,2].plot(alld[:,1],'.')
        # FigsAxes[0,2].set_xlabel('scatter index')
        # FigsAxes[0,2].set_ylabel('d (m)')
        # FigsAxes[1,2].set_xlabel('scatter index')
        # FigsAxes[1,2].set_ylabel('dr (m/s)')
    
        # ssp.utils.force_zeroDoppler_4Simulation(path_d_drate_amp)
        Signals = ssp.integratedSensorSuite.SensorsSignalGeneration_frame(path_d_drate_amp)
        ssp.integratedSensorSuite.SensorsSignalProccessing_Chain_RangeProfile_RangeDoppler_AngleDoppler(Signals,FigsAxes,fig)
        ssp.utils.increaseCurrentFrame()
    # plt.ioff() 
    plt.show()
    
def processing_4():
    ssp.utils.trimUserInputs() 
    ssp.config.restart()
    fig, FigsAxes = plt.subplots(2,2)
    # FigsAxes[1] = fig.add_subplot(1, 2, 2, projection='3d')
    while ssp.config.run():
        path_d_drate_amp = ssp.raytracing.Path_RayTracing_frame()
        fig.suptitle(f'Frame: {ssp.config.CurrentFrame}')
        Signals = ssp.integratedSensorSuite.SensorsSignalGeneration_frame(path_d_drate_amp)
        ssp.integratedSensorSuite.SensorsSignalProccessing_Chain_RangeProfile_RangeDoppler_AngleDoppler_TruckHuman_Detection(Signals,FigsAxes,fig)
        ssp.utils.increaseCurrentFrame()
        print(f'Frame: {ssp.config.CurrentFrame}')
    # plt.ioff() 
    plt.show()

def processing_3():
    ssp.utils.trimUserInputs() 
    ssp.config.restart()
    fig, FigsAxes = plt.subplots(1,2)
    # FigsAxes[1] = fig.add_subplot(1, 2, 2, projection='3d')
    while ssp.config.run():
        path_d_drate_amp = ssp.raytracing.Path_RayTracing_frame()
        # fig2, FigsAxes2 = plt.subplots(len(ssp.lastSuite_Position[0]['Radar'][0]['TX-Position']),
        #                                len(ssp.lastSuite_Position[0]['Radar'][0]['RX-Position']))
        # for itx in range(len(ssp.lastSuite_Position[0]['Radar'][0]['TX-Position'])):
        #     for irx in range(len(ssp.lastSuite_Position[0]['Radar'][0]['RX-Position'])):
        #         print("Test",ssp.config.CurrentFrame,itx,irx,(path_d_drate_amp[0][0][irx][0][0][itx]))
        #         if itx+irx==0:
        #             FigsAxes2[itx,irx].plot([x[0] for x in path_d_drate_amp[0][0][irx][0][0][itx]],[x[2] for x in path_d_drate_amp[0][0][irx][0][0][itx]],'.')
        #         else:
        #             FigsAxes2[itx,irx].plot([x[0] for x in path_d_drate_amp[0][0][irx][0][0][itx]],[x[2] for x in path_d_drate_amp[0][0][irx][0][0][itx]],'.')
        # plt.show()
        fig.suptitle(f'Frame: {ssp.config.CurrentFrame}')
        # FigsAxes[0,2].cla()
        Signals = ssp.integratedSensorSuite.SensorsSignalGeneration_frame(path_d_drate_amp)
        ssp.integratedSensorSuite.SensorsSignalProccessing_Chain_RangeProfile_RangeDoppler_AngleDoppler_TruckHuman(Signals,FigsAxes,fig)
        ssp.utils.increaseCurrentFrame()
        print(f'Frame: {ssp.config.CurrentFrame}')
    # plt.ioff() 
    plt.show()
    
def processing_2():
    ssp.utils.trimUserInputs() 
    ssp.config.restart()
    ssp.config.setDopplerProcessingMethod_FFT_Winv(1)
    fig, FigsAxes = plt.subplots(2,3)
    while ssp.config.run():
        path_d_drate_amp = ssp.raytracing.Path_RayTracing_frame()
        # ssp.utils.force_zeroDoppler_4Simulation(path_d_drate_amp)
        # Channel_d_fd_amp = ssp.visualization.visualize_radar_path_d_drate_amp(path_d_drate_amp,1)
        radar_path_d_drate_amp(path_d_drate_amp,[fig,FigsAxes])
        ssp.utils.increaseCurrentFrame()
    # plt.show()
    

def radar_path_d_drate_amp(path_d_drate_amp,FigsAxes):
    alld = []
    m = path_d_drate_amp[0][0][0][0][0][0][0][3]
    for itx in range(len(path_d_drate_amp[0][0][0][0][0])):
        for irx in range(len(path_d_drate_amp[0][0])):
            for d_drate_amp in path_d_drate_amp[0][0][irx][0][0][itx]:
                if d_drate_amp[3]==m:
                    alld.append(d_drate_amp[0:3])
    alld = np.array(alld)
    FigsAxes[1][0,0].plot(alld[:,0])
    FigsAxes[1][0,1].plot(alld[:,1])
    FigsAxes[1][1,0].plot(alld[:,2])
    FigsAxes[1][1,1].plot(np.diff(alld[:,0]))
    FigsAxes[1][1,2].plot(np.diff(np.diff(alld[:,0])))
    image=ssp.visualization.captureFig(FigsAxes[0])
    image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
    cv2.imshow('d drate ', image)
    if cv2.waitKey(50) & 0xFF == ord('q'):
        return

    
    # plt.draw() 
    # plt.pause(0.1)
def rcxchain(st):
    if st=="Hand Gestures":
        1

def raytracing_test():
    ssp.utils.trimUserInputs() 
    ssp.config.restart()
    ssp.config.RayTracing_ReflectionPointEpsilon=1e-4
    empty_obj = bpy.data.objects.get("Plot Rays")
    if empty_obj:
        # Collect all children recursively
        def collect_children_recursive(obj):
            children = list(obj.children)
            for child in obj.children:
                children.extend(collect_children_recursive(child))
            return children

        # Get all descendants of the empty
        descendants = collect_children_recursive(empty_obj)

        # Deselect all objects first
        bpy.ops.object.select_all(action='DESELECT')

        # Select the empty and all its descendants
        empty_obj.select_set(True)
        for child in descendants:
            child.select_set(True)

        # Delete the selected objects
        bpy.ops.object.delete()
    
    bpy.ops.object.empty_add(type='PLAIN_AXES', align='WORLD', location=(0, 0, 0), rotation=(0, 0, 0), scale=(1, 1, 1))
    allrays = bpy.context.object
    allrays.name = f'Plot Rays'
    while ssp.config.run():    
        bpy.ops.object.empty_add(type='PLAIN_AXES', align='WORLD', location=(0, 0, 0), rotation=(0, 0, 0), scale=(1, 1, 1))
        framerays = bpy.context.object
        framerays.name = f'Plot Rays Frame {ssp.config.getCurrentFrame()}'
        framerays.parent=allrays 
        path_d_drate_amp = ssp.raytracing.Path_RayTracing_frame()
        path_index = 1
        for isrx,suiteRX_d_drate_amp in path_d_drate_amp.items():
            for irrx,radarRX_d_drate_amp in suiteRX_d_drate_amp.items():
                for irx,RX_d_drate_amp in radarRX_d_drate_amp.items():
                    for istx,suiteTX_d_drate_amp in RX_d_drate_amp.items():
                        for irtx,radarTX_d_drate_amp in suiteTX_d_drate_amp.items():
                            for itx,TX_d_drate_amp in radarTX_d_drate_amp.items():
                                for ip,p in enumerate(TX_d_drate_amp):
                                    d,dr,a,m=p
                                    print(path_index,len(m),m,d)
                                    tx = ssp.lastSuite_Position[istx]['Radar'][irtx]['TX-Position'][itx] 
                                    rx = ssp.lastSuite_Position[isrx]['Radar'][irrx]['RX-Position'][irx]
                                    v = [tx]
                                    for mi in m:
                                        if isinstance(mi, list) and len(mi) == 2 and isinstance(mi[1], Vector):
                                            v.append(mi[1])
                                        if isinstance(mi, str):
                                            _, is0, ir0, iris0 = mi.split('_')
                                            is0 = int(is0)
                                            ir0 = int(ir0)
                                            iris0 = int(iris0)
                                            v.append(ssp.lastSuite_Position[is0]['RIS'][ir0]['Position'][iris0])
                                    v.append(rx)
                                    ray = ssp.visualization.plot_continuous_curve(v,framerays,curve_name=f'path {path_index} Frame {ssp.config.CurrentFrame}')
                                    ray["Ray Tracing 20log Amp (dB)"]=float(20*np.log10(a))
                                    ray["Ray Tracing distance"]=float(d)
                                    ray["Ray Tracing doppler * f0"]=float(dr)
                                    ray["Middle Number"]=len(m)
                                    path_index+=1
                                    
        ssp.utils.increaseCurrentFrame()
        break

def processing_HandGesture1642():
    ssp.utils.trimUserInputs()  
    ssp.config.restart()
    ssp.config.DopplerProcessingMethod_FFT_Winv = (True)

    ssp.config.directReceivefromTX =  False 
    ssp.config.RadarRX_only_fromscatters_itsTX = True
    ssp.config.RadarRX_only_fromitsTX = True
    ssp.config.Radar_TX_RX_isolation = True

    ssp.utils.useCUDA(True)

    # ssp.utils.set_RayTracing_light()
    ssp.utils.set_RayTracing_balanced()
    # ssp.utils.set_RayTracing_advanced_intense()
    fig, FigsAxes = plt.subplots(2,3)
    FigsAxes[1, 2] = fig.add_subplot(2, 3, 6, projection='3d')
    
    all_pcs = []
    # Real-time processing loop
    while ssp.config.run():
            
        # Initialize 3D plot
        # fig = plt.figure()
        # ax = fig.add_subplot(111, projection='3d')
        # Generate data for the current frame
        path_d_drate_amp = ssp.raytracing.Path_RayTracing_frame()
        Signals = ssp.integratedSensorSuite.SensorsSignalGeneration_frame(path_d_drate_amp)
        pcs = []  # Collect point clouds for the current frame

        # Process each radar's signal
        for isuite, radarSpecifications in enumerate(ssp.RadarSpecifications):
            for iradar, specifications in enumerate(radarSpecifications):
                for XRadar, timeX in Signals[isuite]['radars'][iradar]:
                    # Perform range processing
                    X_fft_fast, d_fft = ssp.radar.utils.rangeprocessing(XRadar, specifications)

                    # Perform Doppler processing
                    rangeDopplerTXRX, f_Doppler = ssp.radar.utils.dopplerprocessing_mimodemodulation(X_fft_fast, specifications)

                    # Clear and update the first plot: ADC Samples vs. ADC Output Level
                    FigsAxes[0, 0].clear()
                    FigsAxes[0, 0].plot(np.real(XRadar[:, 0, 0]), label="Real Part")
                    FigsAxes[0, 0].plot(np.imag(XRadar[:, 0, 0]), label="Imaginary Part")
                    # FigsAxes[0, 0].set_title("ADC Samples vs. ADC Output Level")
                    FigsAxes[0, 0].set_xlabel("ADC Samples")
                    FigsAxes[0, 0].set_ylabel("ADC Output Level")
                    FigsAxes[0, 0].legend(loc='upper right')

                    # Clear and update the second plot: Range vs. ADC Output Level
                    FigsAxes[0, 1].clear()
                    FigsAxes[0, 1].plot(d_fft, np.abs(X_fft_fast[:, 0, 0]))
                    # FigsAxes[0, 1].set_title("Range vs. ADC Output Level")
                    FigsAxes[0, 1].set_xlabel("Range (m)")
                    FigsAxes[0, 1].set_ylabel("Range Profile")

                    # Clear and update the third plot: Range-Doppler Heatmap
                    FigsAxes[0, 2].clear()
                    im = FigsAxes[0, 2].imshow(np.abs(rangeDopplerTXRX[:, :, 0, 0]),
                                            extent=[f_Doppler[0], f_Doppler[-1], d_fft[0], d_fft[-1]],
                                            aspect='auto', origin='lower')
                    # FigsAxes[0, 2].set_title("Range-Doppler Map")
                    FigsAxes[0, 2].set_xlabel("Doppler Frequency (Hz)")
                    FigsAxes[0, 2].set_ylabel("Range (m)")
                    if not hasattr(FigsAxes[0, 2], "colorbar_added"):  # Add colorbar only once
                        colorbar = fig.colorbar(im, ax=FigsAxes[0, 2], label="Amplitude")
                        FigsAxes[0, 2].colorbar_added = True
                    # CFAR detection
                    detections, cfar_threshold, rangeDoppler4CFAR = ssp.radar.utils.rangedoppler_detection_alpha(rangeDopplerTXRX, specifications)

                    # Plot CFAR detection results
                    FigsAxes[1, 0].clear()
                    im = FigsAxes[1, 0].imshow(rangeDoppler4CFAR, 
                                            extent=[f_Doppler[0], f_Doppler[-1], d_fft[0], d_fft[-1]],
                                            aspect='auto', origin='lower', cmap='viridis')
                    FigsAxes[1, 0].set_xlabel("Doppler Frequency (Hz)")
                    FigsAxes[1, 0].set_ylabel("Range (m)")
                    FigsAxes[1, 0].set_title("CFAR Detection Map")
                    if not hasattr(FigsAxes[1, 0], "colorbar_added"):
                        fig.colorbar(im, ax=FigsAxes[1, 0], label="Detection Amplitude")
                        FigsAxes[1, 0].colorbar_added = True

                    
                    # distance = np.linspace(d_fft[0], d_fft[-1], rangeDoppler4CFAR.shape[0])  # Replace with real distance data
                    # elevation = np.linspace(f_Doppler[0],f_Doppler[-1], rangeDoppler4CFAR.shape[1])  # Replace with real angle data
                    # X, Y = np.meshgrid(elevation, distance)
                    # ax.plot_surface(X, Y, 10*np.log10(rangeDoppler4CFAR), cmap='viridis', alpha=1)
                    # ax.plot_surface(X, Y, 10*np.log10(cfar_threshold)+0, color='yellow', alpha=1)
                    # detected_points = np.where(detections == 1)
                    # ax.scatter(elevation[detected_points[1]], distance[detected_points[0]], 
                    #        10*np.log10(rangeDoppler4CFAR[detected_points]), color='red', s=20, label='Post-CFAR Point Cloud')
                    # plt.show()
                    # Extract detected points
                    detected_points = np.where(detections == 1)

                    # Angle processing and point cloud generation
                    Azimuth_Angles,debug_spectrums = ssp.radar.utils.angleprocessing_capon1D(rangeDopplerTXRX, detections, specifications)
                    pc = ssp.radar.utils.pointscloud(d_fft[detected_points[0]], f_Doppler[detected_points[1]], Azimuth_Angles)
                    pcs.append(pc)
                    
                    FigsAxes[1, 1].clear()
                    for debug_spectrum in debug_spectrums:
                        FigsAxes[1, 1].plot(debug_spectrum)
                        
                    FigsAxes[1, 2].clear()
                    update_pointclouds(all_pcs,FigsAxes[1, 2])
                     
                    plt.draw()
                    plt.pause(0.1)
                    plt.gcf().canvas.flush_events()
        all_pcs.append(pcs)

        print(f'Processed frame = {ssp.config.CurrentFrame}')
        ssp.utils.increaseCurrentFrame()

    plt.legend()

    plt.show()


def update_pointclouds(all_pcs,ax):
    
    x,y,z,v=np.array([]),np.array([]),np.array([]),np.array([])
    for i, pcs in enumerate(all_pcs):
        for j,pc in enumerate(pcs):
            if len(pc)==0:
                continue
            y=np.append(y,pc[:, 1])
            z=np.append(z,pc[:, 0])
            v=np.append(v,pc[:, 2])
            x=np.append(x,np.ones_like(pc[:, 0])*i)
    ind_filter = np.where(np.abs(v)>-70)
    x=x[ind_filter]
    y=y[ind_filter]
    z=z[ind_filter]
    v=v[ind_filter]
    if len(v)==0:
        return
    # ax = fig.add_subplot(111, projection='3d')

    # Normalize the v values for better color mapping
    norm = plt.Normalize(v.min(), v.max())


    # Scatter plot with colors based on v values
    scatter = ax.scatter(x, y, z, c=v, cmap='viridis', norm=norm)

    # Add a colorbar to interpret the color scale
    # cbar = fig.colorbar(scatter, ax=ax, label='v value')

    # ssp.utils.set_axes_equal(ax)    
    ax.set_ylabel('X')
    ax.set_xlabel('Frame')
    ax.set_zlabel('Y')
    ax.set_xlim([0,30])
    


import bpy 
import sensingsp as ssp
import numpy as np
from mathutils import Vector

def add_radar_string(st):
    suite_planes = ssp.environment.BlenderSuiteFinder().find_suite_planes()
    obj = bpy.context.view_layer.objects.active
    if obj in suite_planes:
        suiteIndex = int(obj.name.split('_')[-1])
        radar_planes = ssp.environment.BlenderSuiteFinder().find_radar_planes(obj)
        radarIndex = max([int(plane.name.split('_')[2]) for plane in radar_planes if plane.parent == obj] or [-1]) + 1
    else:
        suiteIndex = max([int(plane.name.split('_')[-1]) for plane in suite_planes] or [-1]) + 1
        ssp.integratedSensorSuite.define_suite(suiteIndex, location=Vector((0, 0, 0)), rotation=Vector((0, 0, 0)))
        radarIndex = 0  
    if st=='TI AWR 1642 (2TX-4RX)':
        freq = 76e9
        ssp.radar.utils.predefined_array_configs_TI_AWR1642(isuite=suiteIndex, iradar=radarIndex, location=Vector((0, 0,0)), rotation=Vector((np.pi/2,0, -np.pi/2)), f0=freq)
    if st=='TI IWR 6843 (3TX-4RX)':
        freq = 76e9
        ssp.radar.utils.predefined_array_configs_TI_IWR6843(isuite=suiteIndex, iradar=radarIndex, location=Vector((0, 0,0)), rotation=Vector((np.pi/2,0, -np.pi/2)), f0=freq)
    if st=='TI Cascade AWR 2243 (12TX-16RX)':
        freq = 76e9
        ssp.radar.utils.predefined_array_configs_TI_Cascade_AWR2243(isuite=suiteIndex, iradar=radarIndex, location=Vector((0, 0,0)), rotation=Vector((np.pi/2,0, -np.pi/2)), f0=freq)
    if st=='SISO':
        freq = 76e9
        ssp.radar.utils.predefined_array_configs_SISO(isuite=suiteIndex, iradar=radarIndex, location=Vector((0, 0,0)), rotation=Vector((np.pi/2,0, -np.pi/2)), f0=freq)
    if st=='JSON':
        1
        # ssp.radar.utils.predefined_array_configs_JSON(isuite=suiteIndex, iradar=radarIndex, location=Vector((0, 0,0)), rotation=Vector((np.pi/2,0, -np.pi/2)),file_path="")

def add_sensors_string(st):
    1

def add_radar(context,option):
    freq = float(context.scene.radar_properties.freq)*1e9  # Convert string to float
    if "Simulation Settings" in bpy.data.objects:
        if "RF Frequency (GHz)" not in bpy.data.objects["Simulation Settings"]:
          bpy.data.objects["Simulation Settings"]["RF Frequency (GHz)"]=1
        freq = bpy.data.objects["Simulation Settings"]["RF Frequency (GHz)"]*1e9
          
    
    if freq<=0:
        freq=1
    # finder = BlenderSuiteFinder()
    # suite_manager = BlenderRadarSuiteManager()
    suite_planes = ssp.environment.BlenderSuiteFinder().find_suite_planes()
    obj = bpy.context.view_layer.objects.active
    if obj in suite_planes:
        suiteIndex = int(obj.name.split('_')[-1])
        radar_planes = ssp.environment.BlenderSuiteFinder().find_radar_planes(obj)
        radarIndex = max([int(plane.name.split('_')[2]) for plane in radar_planes if plane.parent == obj] or [-1]) + 1
        if option=="Cascade":
            ssp.radar.utils.predefined_array_configs_TI_Cascade_AWR2243(isuite=suiteIndex, iradar=radarIndex, location=Vector((0, 0,0)), rotation=Vector((np.pi/2,-np.pi/2, -np.pi/2)), f0=freq)
        if option=="6843":
            ssp.radar.utils.predefined_array_configs_TI_IWR6843(isuite=suiteIndex, iradar=radarIndex, location=Vector((0, 0,0)), rotation=Vector((np.pi/2,-np.pi/2, -np.pi/2)), f0=freq)
        if option=="SISO":
            ssp.radar.utils.predefined_array_configs_SISO(isuite=suiteIndex, iradar=radarIndex, location=Vector((0, 0,0)), rotation=Vector((np.pi/2,-np.pi/2, -np.pi/2)), f0=freq)
        if option=="awr1642":
            ssp.radar.utils.predefined_array_configs_TI_AWR1642(isuite=suiteIndex, iradar=radarIndex, location=Vector((0, 0,0)), rotation=Vector((np.pi/2,0, -np.pi/2)), f0=freq)
        if option=="JSON":
            ssp.radar.utils.predefined_array_configs_JSON(isuite=suiteIndex, iradar=radarIndex, location=Vector((0, 0,0)), rotation=Vector((np.pi/2,-np.pi/2, -np.pi/2)), f0=freq)
    else:
        suiteIndex = max([int(plane.name.split('_')[-1]) for plane in suite_planes] or [-1]) + 1
        ssp.integratedSensorSuite.define_suite(suiteIndex, location=Vector((0, 0, 0)), rotation=Vector((0, 0, 0)))
        if option=="Cascade":
            ssp.radar.utils.predefined_array_configs_TI_Cascade_AWR2243(isuite=suiteIndex, iradar=0, location=Vector((0, 0,0)), rotation=Vector((np.pi/2,-np.pi/2, -np.pi/2)), f0=freq)
        if option=="6843":
            ssp.radar.utils.predefined_array_configs_TI_IWR6843(isuite=suiteIndex, iradar=0, location=Vector((0, 0,0)), rotation=Vector((np.pi/2,-np.pi/2, -np.pi/2)), f0=freq)
        if option=="SISO":
            ssp.radar.utils.predefined_array_configs_SISO(isuite=suiteIndex, iradar=0, location=Vector((0, 0,0)), rotation=Vector((np.pi/2,-np.pi/2, -np.pi/2)), f0=freq)
        if option=="awr1642":
            ssp.radar.utils.predefined_array_configs_TI_AWR1642(isuite=suiteIndex, iradar=0, location=Vector((0, 0,0)), rotation=Vector((np.pi/2,0, -np.pi/2)), f0=freq)
        if option=="JSON":
            ssp.radar.utils.predefined_array_configs_JSON(isuite=suiteIndex, iradar=0, location=Vector((0, 0,0)), rotation=Vector((np.pi/2,-np.pi/2, -np.pi/2)), f0=freq)

def add_radar_Json(file_path):
    if not file_path.lower().endswith('.json'):
        print("Error: The selected file is not a .json file.")
        return None
    
    suite_planes = ssp.environment.BlenderSuiteFinder().find_suite_planes()
    obj = bpy.context.view_layer.objects.active
    if obj in suite_planes:
        suiteIndex = int(obj.name.split('_')[-1])
        radar_planes = ssp.environment.BlenderSuiteFinder().find_radar_planes(obj)
        radarIndex = max([int(plane.name.split('_')[2]) for plane in radar_planes if plane.parent == obj] or [-1]) + 1
        ssp.radar.utils.predefined_array_configs_JSON(isuite=suiteIndex, iradar=radarIndex, location=Vector((0, 0,0)), rotation=Vector((np.pi/2,-np.pi/2, -np.pi/2)), file_path=file_path)
        
    else:
        suiteIndex = max([int(plane.name.split('_')[-1]) for plane in suite_planes] or [-1]) + 1
        ssp.integratedSensorSuite.define_suite(suiteIndex, location=Vector((0, 0, 0)), rotation=Vector((0, 0, 0)))
        ssp.radar.utils.predefined_array_configs_JSON(isuite=suiteIndex, iradar=0, location=Vector((0, 0,0)), rotation=Vector((np.pi/2,-np.pi/2, -np.pi/2)), file_path=file_path)
        
    
def predefine_cube_cascade():
    ssp.utils.delete_all_objects()
    ssp.environment.add_cube(location=Vector((5, 0, 0)), direction=Vector((1, 0, 0)), scale=Vector((.1, .1, .1)), subdivision=0)
    ssp.integratedSensorSuite.define_suite(0, location=Vector((0, 0, 0)), rotation=Vector((0, 0, 0)))
    ssp.radar.utils.predefined_array_configs_TI_Cascade_AWR2243(isuite=0, iradar=0, location=Vector((0, 0,0)), rotation=Vector((np.pi/2,-np.pi/2, -np.pi/2)), f0=70e9)
    ssp.utils.set_frame_start_end(start=1,end=2)

def predefine_movingcube_6843():
    ssp.utils.delete_all_objects()
    ssp.utils.define_settings()
    cube = ssp.environment.add_cube(location=Vector((0, 0, 0)), direction=Vector((1, 0, 0)), scale=Vector((.1, .1, .1)), subdivision=0)
    cube["RCS0"]=.0001
    cube.location = (3,0,0)
    cube.keyframe_insert(data_path="location", frame=1)
    cube.location = (3, 3,0)
    cube.keyframe_insert(data_path="location", frame=30)
    cube.location = (3, -3,0)
    cube.keyframe_insert(data_path="location", frame=100)
    for fcurve in cube.animation_data.action.fcurves:
        for keyframe in fcurve.keyframe_points:
            keyframe.interpolation = 'LINEAR'
    ssp.integratedSensorSuite.define_suite(0, location=Vector((0, 0, 0)), rotation=Vector((0, 0, 0)))
    ssp.radar.utils.predefined_array_configs_TI_IWR6843(isuite=0, iradar=0, location=Vector((0, 0,0)), rotation=Vector((np.pi/2,0, -np.pi/2)), f0=70e9)
    ssp.utils.set_frame_start_end(start=1,end=100)


def add_camera(context,option):
    suite_planes = ssp.environment.BlenderSuiteFinder().find_suite_planes()
    obj = bpy.context.view_layer.objects.active
    if obj in suite_planes:
        suiteIndex = int(obj.name.split('_')[-1])
        camera_planes = ssp.environment.BlenderSuiteFinder().find_camera_planes(obj)
        cameraIndex = max([int(plane.name.split('_')[2]) for plane in camera_planes if plane.parent == obj] or [-1]) + 1
        ssp.camera.utils.add_camera(isuite=suiteIndex, icamera=cameraIndex, location=Vector((0, 0,0)), rotation=Vector((np.pi/2,-np.pi/2, -np.pi/2)))
        
    else:
        suiteIndex = max([int(plane.name.split('_')[-1]) for plane in suite_planes] or [-1]) + 1
        ssp.integratedSensorSuite.define_suite(suiteIndex, location=Vector((0, 0, 0)), rotation=Vector((0, 0, 0)))
        ssp.camera.utils.add_camera(isuite=suiteIndex, icamera=0, location=Vector((0, 0,0)), rotation=Vector((np.pi/2,-np.pi/2, -np.pi/2)))

def add_lidar(context,option):
    suite_planes = ssp.environment.BlenderSuiteFinder().find_suite_planes()
    obj = bpy.context.view_layer.objects.active
    if obj in suite_planes:
        suiteIndex = int(obj.name.split('_')[-1])
        lidar_planes = ssp.environment.BlenderSuiteFinder().find_lidar_planes(obj)
        lidarIndex = max([int(plane.name.split('_')[2]) for plane in lidar_planes if plane.parent == obj] or [-1]) + 1
        ssp.lidar.utils.add_lidar(isuite=suiteIndex, ilidar=lidarIndex, location=Vector((0, 0,0)), rotation=Vector((np.pi/2,-np.pi/2, -np.pi/2)))
        
    else:
        suiteIndex = max([int(plane.name.split('_')[-1]) for plane in suite_planes] or [-1]) + 1
        ssp.integratedSensorSuite.define_suite(suiteIndex, location=Vector((0, 0, 0)), rotation=Vector((0, 0, 0)))
        ssp.lidar.utils.add_lidar(isuite=suiteIndex, ilidar=0, location=Vector((0, 0,0)), rotation=Vector((np.pi/2,-np.pi/2, -np.pi/2)))
        
def add_ris(context,option):
    suite_planes = ssp.environment.BlenderSuiteFinder().find_suite_planes()
    obj = bpy.context.view_layer.objects.active
    if obj in suite_planes:
        suiteIndex = int(obj.name.split('_')[-1])
        ris_planes = ssp.environment.BlenderSuiteFinder().find_ris_planes(obj)
        risIndex = max([int(plane.name.split('_')[2]) for plane in ris_planes if plane.parent == obj] or [-1]) + 1
        ssp.ris.utils.add_ris(isuite=suiteIndex, iris=risIndex, location=Vector((0, 0,0)), rotation=Vector((np.pi/2,-np.pi/2, -np.pi/2)))
        
    else:
        suiteIndex = max([int(plane.name.split('_')[-1]) for plane in suite_planes] or [-1]) + 1
        ssp.integratedSensorSuite.define_suite(suiteIndex, location=Vector((0, 0, 0)), rotation=Vector((0, 0, 0)))
        ssp.ris.utils.add_ris(isuite=suiteIndex, iris=0, location=Vector((0, 0,0)), rotation=Vector((np.pi/2,-np.pi/2, -np.pi/2)))
        
def add_JRC(context,option):
    suite_planes = ssp.environment.BlenderSuiteFinder().find_suite_planes()
    obj = bpy.context.view_layer.objects.active
    if obj in suite_planes:
        suiteIndex = int(obj.name.split('_')[-1])
        # ris_planes = ssp.environment.BlenderSuiteFinder().find_ris_planes(obj)
        # risIndex = max([int(plane.name.split('_')[2]) for plane in ris_planes if plane.parent == obj] or [-1]) + 1
        # ssp.ris.utils.add_ris(isuite=suiteIndex, iris=risIndex, location=Vector((0, 0,0)), rotation=Vector((np.pi/2,-np.pi/2, -np.pi/2)))
        
    else:
        suiteIndex = max([int(plane.name.split('_')[-1]) for plane in suite_planes] or [-1]) + 1
        # ssp.integratedSensorSuite.define_suite(suiteIndex, location=Vector((0, 0, 0)), rotation=Vector((0, 0, 0)))
        # ssp.ris.utils.add_ris(isuite=suiteIndex, iris=0, location=Vector((0, 0,0)), rotation=Vector((np.pi/2,-np.pi/2, -np.pi/2)))
        
def add_Probe(context,option):
    suite_planes = ssp.environment.BlenderSuiteFinder().find_suite_planes()
    obj = bpy.context.view_layer.objects.active
    if obj in suite_planes:
        suiteIndex = int(obj.name.split('_')[-1])
        probe_planes = ssp.environment.BlenderSuiteFinder().find_probe_planes(obj)
        probeIndex = max([int(plane.name.split('_')[2]) for plane in probe_planes if plane.parent == obj] or [-1]) + 1
        ssp.probe.utils.add_probe(isuite=suiteIndex, iprobe=probeIndex, location=Vector((0, 0,0)), rotation=Vector((np.pi/2,-np.pi/2, -np.pi/2)))
        
    else:
        suiteIndex = max([int(plane.name.split('_')[-1]) for plane in suite_planes] or [-1]) + 1
        ssp.integratedSensorSuite.define_suite(suiteIndex, location=Vector((0, 0, 0)), rotation=Vector((0, 0, 0)))
        ssp.probe.utils.add_probe(isuite=suiteIndex, iprobe=0, location=Vector((0, 0,0)), rotation=Vector((np.pi/2,-np.pi/2, -np.pi/2)))
        


import bpy

# Function to create a material with the car texture
def create_car_material(image_path):
    # Create a new material
    mat = bpy.data.materials.new(name="CarMaterial")
    mat.use_nodes = True
    
    # Get the material's node tree
    bsdf = mat.node_tree.nodes["Principled BSDF"]
    tex_image = mat.node_tree.nodes.new('ShaderNodeTexImage')
    
    # Load the texture
    tex_image.image = bpy.data.images.load(image_path)
    
    # Connect the texture to the Base Color of the shader
    mat.node_tree.links.new(bsdf.inputs['Base Color'], tex_image.outputs['Color'])
    
    return mat

# Function to create a wheel
def create_wheel(location):
    bpy.ops.mesh.primitive_cylinder_add(radius=0.4, depth=0.2, location=location)
    wheel = bpy.context.object
    wheel.rotation_euler = (1.5708, 0, 0)
    bpy.ops.object.shade_smooth()
    return wheel

# Function to create the car body
def create_body(location):
    bpy.ops.mesh.primitive_cube_add(size=2.5, location=location)
    body = bpy.context.object
    body.scale[2] = 0.5
    body.scale[1] = 1.0
    body.scale[0] = 2.0
    return body

# Function to create the car roof
def create_roof(location):
    bpy.ops.mesh.primitive_cube_add(size=1.5, location=location)
    roof = bpy.context.object
    roof.scale[2] = 0.5
    roof.scale[1] = 0.7
    roof.scale[0] = 1.2
    return roof

# Function to create the entire car
def create_car(location=(0, 0, 0), material=None):
    # Create car body
    body = create_body((location[0], location[1], location[2] + 1))
    
    # Apply material to the body
    if material:
        if body.data.materials:
            body.data.materials[0] = material
        else:
            body.data.materials.append(material)
    
    # Create car roof
    roof = create_roof((location[0], location[1], location[2] + 2))
    
    # Apply material to the roof
    if material:
        if roof.data.materials:
            roof.data.materials[0] = material
        else:
            roof.data.materials.append(material)
    
    # Create wheels
    wheel1 = create_wheel((location[0] + 1.5, location[1] + 1.2, location[2] + 0.5))
    wheel2 = create_wheel((location[0] + 1.5, location[1] - 1.2, location[2] + 0.5))
    wheel3 = create_wheel((location[0] - 1.5, location[1] + 1.2, location[2] + 0.5))
    wheel4 = create_wheel((location[0] - 1.5, location[1] - 1.2, location[2] + 0.5))
    
    # Combine all parts into a single object
    bpy.ops.object.select_all(action='DESELECT')
    body.select_set(True)
    roof.select_set(True)
    wheel1.select_set(True)
    wheel2.select_set(True)
    wheel3.select_set(True)
    wheel4.select_set(True)
    bpy.ops.object.join()
    
    # Move the entire car to the specified location (optional)
    bpy.context.object.location = location
    
    # Set the name of the car object
    bpy.context.object.name = "SimpleCar"
    
    # Center the origin of the car
    bpy.ops.object.origin_set(type='ORIGIN_CENTER_OF_MASS', center='BOUNDS')
    
    # Return the car object
    return bpy.context.object

# # Path to your car texture image
# image_path = "/path/to/your/car_texture_image.jpg"

# # Create the car material
# car_material = create_car_material(image_path)

# # Example usage: create a car at the origin with the car material
# car = create_car(location=(0, 0, 0), material=car_material)



import sensingsp as ssp
from ..environment import BlenderSuiteFinder
from ..radar.utils import MIMO_Functions
import os
import sys
import bpy
import numpy as np
from mathutils import Vector
import bmesh
import subprocess
import platform
import time
import matplotlib.pyplot as plt
import cv2

def set_RayTracing_advanced_intense():  
  bpy.data.objects["Simulation Settings"]["do RayTracing LOS"] = True
  bpy.data.objects["Simulation Settings"]["do RayTracing Simple"] = False

def setRadar_oneCPI_in_oneFrame(radar):
  radar['continuousCPIsTrue_oneCPIpeerFrameFalse']=False
def setRadar_multipleCPI_in_oneFrame(radar):
  radar['continuousCPIsTrue_oneCPIpeerFrameFalse']=True

def set_RayTracing_balanced():  
  bpy.data.objects["Simulation Settings"]["do RayTracing LOS"] = True
  bpy.data.objects["Simulation Settings"]["do RayTracing Simple"] = True
  
def set_RayTracing_light():  
  bpy.data.objects["Simulation Settings"]["do RayTracing LOS"] = False
  bpy.data.objects["Simulation Settings"]["do RayTracing Simple"] = False

def set_raytracing_bounce(N):
  bpy.data.objects["Simulation Settings"]["Bounce Number"] = N

def set_debugSettings(x):
  bpy.data.objects["Simulation Settings"]["debug Settings"] = x
def get_debugSettings():
  return bpy.data.objects["Simulation Settings"]["debug Settings"] 
        
def tic():
  start_time = time.perf_counter()
  return start_time
def toc(start_time,s=''):
  elapsed_time = time.perf_counter() - start_time
  print(f"Elapsed time: {elapsed_time} seconds {s}")

def pointCloud_RangeAzimuthElevation_THR(RangeAzimuthElevation=None,rangeResolution_maxUnambigiousRange=1,THR=0.05):
  if RangeAzimuthElevation is None:
      return
  X=np.abs(RangeAzimuthElevation)
  max_value = np.max(X)
  fig = plt.figure(figsize=(10, 6))
  ax = fig.add_subplot(111, projection='3d')
  x,y,z,values_above_threshold=[],[],[],[]
  AzimuthElevationRangeindexofMaxValue = np.argmax(X, axis=0)
  for iaz in range(AzimuthElevationRangeindexofMaxValue.shape[0]):
    az = np.arcsin(-1+2*iaz/AzimuthElevationRangeindexofMaxValue.shape[0])
    for iel in range(AzimuthElevationRangeindexofMaxValue.shape[1]):
      el = np.arcsin(-1+2*iel/AzimuthElevationRangeindexofMaxValue.shape[1])
      ir = AzimuthElevationRangeindexofMaxValue[iaz,iel]
      # if X[ir,iaz,iel] > 500*np.mean(X[:,iaz,iel]):
      if X[ir,iaz,iel] > max_value*THR:
        x0,y0,z0=sph2cart(ir*rangeResolution_maxUnambigiousRange, az, el)
        x.append(x0)
        y.append(y0)
        z.append(z0)
        values_above_threshold.append(X[ir,iaz,iel])
  sc = ax.scatter(x, y, z, c=values_above_threshold, cmap='viridis', marker='o')
  plt.colorbar(sc)
  # Make the axes limits equal
  max_range = np.array([max(x)-min(x), max(y)-min(y), max(z)-min(z)]).max() / 2.0
  
  mean_x = np.mean([max(x), min(x)])
  mean_y = np.mean([max(y), min(y)])
  mean_z = np.mean([max(z), min(z)])
  
  ax.set_xlim(mean_x - max_range, mean_x + max_range)
  ax.set_ylim(mean_y - max_range, mean_y + max_range)
  ax.set_zlim(mean_z - max_range, mean_z + max_range)
  plt.show()
def sph2cart(r, azimuth, elevation):
    x = r * np.cos(elevation) * np.cos(azimuth)
    y = r * np.cos(elevation) * np.sin(azimuth)
    z = r * np.sin(elevation)
    return x, y, z
def imshow(X=None, xlabel='', ylabel='', title=''):
    if X is None:
        return
    plt.figure(figsize=(10, 6))
    plt.imshow(X, extent=None, aspect='auto')
    if xlabel:
        plt.xlabel(xlabel)
    if ylabel:
        plt.ylabel(ylabel)
    if title:
        plt.title(title)
    plt.show()
                

def plot(x=None, y=None, option=1, xlabel='', ylabel='', title=''):
    # Initialize x and y to empty lists if None is provided
    if x is None:
        x = []
    if y is None:
        y = []
    
    # Create the plot
    plt.figure(figsize=(10, 6))
    
    # Check if y is empty or not
    if not y:  # If y is empty, plot only x
        plt.plot(x)
    else:
        if len(x) == len(y):  # If x and y have the same length, plot x vs y
            plt.plot(x, y)
    
    # Optionally add labels and title
    if xlabel:
        plt.xlabel(xlabel)
    if ylabel:
        plt.ylabel(ylabel)
    if title:
        plt.title(title)
    
    # Add grid and show plot
    plt.grid(True)
    plt.show()
def plot_text(s):
  if len(s)==0:
    return
  plt.figure(figsize=(10, 6))
  for i, a in enumerate(s):
      plt.text(0.01, 1-i * 0.1, a, fontsize=12)  # Adjust y-coordinate to start from the top.
  plt.axis('off')                               # Hides the axis for a cleaner look.
  plt.tight_layout()
  plt.show()
def delete_all_objects():
    view_layer = bpy.context.view_layer
    bpy.ops.object.select_all(action='DESELECT')
    for obj in bpy.data.objects:
        if obj.name in view_layer.objects.keys():
            view_layer.objects.active = obj
            obj.select_set(True)
    if "Simulation Settings" in bpy.data.objects:
      bpy.data.objects["Simulation Settings"].select_set(False)
    bpy.ops.object.delete()
def save_Blender(folder="",file="save_frompython.blend"):
  if folder == "":
    folder = ssp.config.temp_folder

  bpy.ops.wm.save_as_mainfile(filepath = os.path.join(folder, file))
  print(os.path.join(folder, file))
  
def getRadarSpecs():
  return ssp.RadarSpecifications

def useCUDA(value=True):
    bpy.data.objects["Simulation Settings"]["CUDA SignalGeneration Enabled"] = value
    
def set_frame_start_end(start=1,end=2):
    
    bpy.context.scene.frame_start   =   start
    bpy.context.scene.frame_end     =   end
def increaseCurrentFrame(step=1):
    ssp.config.CurrentFrame += step
    c= sys.executable.lower()
    if "blender" in c:
      bpy.context.view_layer.update()
      bpy.ops.wm.redraw_timer(type='DRAW_WIN_SWAP', iterations=1)

def total_paths(path_d_drate_amp):
  N=0
  for isrx,suiteRX_d_drate_amp in path_d_drate_amp.items():
    for irrx,radarRX_d_drate_amp in suiteRX_d_drate_amp.items():
      for irx,RX_d_drate_amp in radarRX_d_drate_amp.items():
        for istx,suiteTX_d_drate_amp in RX_d_drate_amp.items():
          for irtx,radarTX_d_drate_amp in suiteTX_d_drate_amp.items():
            for itx,TX_d_drate_amp in radarTX_d_drate_amp.items():
              N+=len(TX_d_drate_amp)
  return N
def force_zeroDoppler_4Simulation(path_d_drate_amp):
  d=[]
  for isrx,suiteRX_d_drate_amp in path_d_drate_amp.items():
    for irrx,radarRX_d_drate_amp in suiteRX_d_drate_amp.items():
      for irx,RX_d_drate_amp in radarRX_d_drate_amp.items():
        for istx,suiteTX_d_drate_amp in RX_d_drate_amp.items():
          for irtx,radarTX_d_drate_amp in suiteTX_d_drate_amp.items():
            for itx,TX_d_drate_amp in radarTX_d_drate_amp.items():
              for d_drate_amp in TX_d_drate_amp:
                d_drate_amp[1]=0
                d.append(d_drate_amp[0])
  return np.array(d)
def zeroDopplerCancellation_4Simulation(path_d_drate_amp,attenuation_dB=80):
  attenuation = 10**(-attenuation_dB/20)
  for isrx,suiteRX_d_drate_amp in path_d_drate_amp.items():
    for irrx,radarRX_d_drate_amp in suiteRX_d_drate_amp.items():
      for irx,RX_d_drate_amp in radarRX_d_drate_amp.items():
        for istx,suiteTX_d_drate_amp in RX_d_drate_amp.items():
          for irtx,radarTX_d_drate_amp in suiteTX_d_drate_amp.items():
            for itx,TX_d_drate_amp in radarTX_d_drate_amp.items():
              for d_drate_amp in TX_d_drate_amp:
                if d_drate_amp[1]==0:
                  d_drate_amp[2]*=attenuation

def trimUserInputs():
    current_working_directory = os.getcwd()
    if "Simulation Settings" in bpy.data.objects:
        sim_axes = bpy.data.objects["Simulation Settings"]
        RenderBlenderFrames = bpy.data.objects["Simulation Settings"]["Render Blender Frames"]
        video_directory = bpy.data.objects["Simulation Settings"]["Video Directory"]
        open_output_folder = bpy.data.objects["Simulation Settings"]["Open Output Folder"]
    else:
        RenderBlenderFrames = True
        video_directory = current_working_directory
        open_output_folder = True
    RadarSpecifications = []
    suite_information = BlenderSuiteFinder().find_suite_information()
    
    ssp.suite_information = suite_information
    # suite_information= finder.find_suite_information()
    mimo_Functions = MIMO_Functions()
    for isuite,suiteobject in enumerate(suite_information):
      radarSpecifications=[]
      for iradar,radarobject in enumerate(suiteobject['Radar']):
        specifications={}

        # empty["Transmit_Power_dBm"] = 12
        # empty["Center_Frequency_GHz"] = f0/1e9
        # empty['Fs_MHz']=5
        # empty['FMCW_ChirpTime_us'] = 60 automatically set to = N_ADC * Ts

        specifications['PRI']=radarobject['GeneralRadarSpec_Object']['PRI_us']*1e-6
        specifications['Ts']=1e-6/radarobject['GeneralRadarSpec_Object']['Fs_MHz']
        specifications['NPulse'] = radarobject['GeneralRadarSpec_Object']['NPulse']
        specifications['N_ADC']  = radarobject['GeneralRadarSpec_Object']['N_ADC']
        
        specifications['Lambda']=ssp.constants.LightSpeed/radarobject['GeneralRadarSpec_Object']["Center_Frequency_GHz"]/1e9
        specifications['RadarMode']=radarobject['GeneralRadarSpec_Object']['RadarMode']
        specifications['PulseWaveform']=radarobject['GeneralRadarSpec_Object']['PulseWaveform']
        # specifications['FMCW_Bandwidth']=radarobject['GeneralRadarSpec_Object']['FMCW_Bandwidth_GHz']*1e9
        specifications['Tempreture_K']=radarobject['GeneralRadarSpec_Object']['Tempreture_K']
        specifications['FMCW_ChirpSlobe'] = radarobject['GeneralRadarSpec_Object']['FMCW_ChirpSlobe_MHz_usec']*1e12
        # specifications['PrecodingMatrix'] = np.eye(len(Suite_Position[isuite]['Radar'][iradar]['TX-Position']),dtype=np.complex128)
        specifications['M_TX'] = len(radarobject['TX'])
        specifications['N_RX'] = len(radarobject['RX'])
        specifications['MIMO_Tech'] = 'TDM'
        specifications['RangeFFT_OverNextP2'] = radarobject['GeneralRadarSpec_Object']['RangeFFT_OverNextP2']
        specifications['Range_Start'] = radarobject['GeneralRadarSpec_Object']['Range_Start']
        specifications['Range_End'] = radarobject['GeneralRadarSpec_Object']['Range_End']
        specifications['CFAR_RD_guard_cells'] = radarobject['GeneralRadarSpec_Object']['CFAR_RD_guard_cells']
        specifications['CFAR_RD_training_cells'] = radarobject['GeneralRadarSpec_Object']['CFAR_RD_training_cells']
        specifications['CFAR_RD_false_alarm_rate'] = radarobject['GeneralRadarSpec_Object']['CFAR_RD_false_alarm_rate']
        specifications['STC_Enabled'] = radarobject['GeneralRadarSpec_Object']['STC_Enabled']
        specifications['MTI_Enabled'] = radarobject['GeneralRadarSpec_Object']['MTI_Enabled']
        specifications['DopplerFFT_OverNextP2'] = radarobject['GeneralRadarSpec_Object']['DopplerFFT_OverNextP2']
        specifications['AzFFT_OverNextP2'] = radarobject['GeneralRadarSpec_Object']['AzFFT_OverNextP2']
        specifications['ElFFT_OverNextP2'] = radarobject['GeneralRadarSpec_Object']['ElFFT_OverNextP2']
        specifications['CFAR_Angle_guard_cells'] = radarobject['GeneralRadarSpec_Object']['CFAR_Angle_guard_cells']
        specifications['CFAR_Angle_training_cells'] = radarobject['GeneralRadarSpec_Object']['CFAR_Angle_training_cells']
        specifications['CFAR_Angle_false_alarm_rate'] = radarobject['GeneralRadarSpec_Object']['CFAR_Angle_false_alarm_rate']
        specifications['CFAR_RD_alpha'] = radarobject['GeneralRadarSpec_Object']['CFAR_RD_alpha']
        specifications['CFAR_Angle_alpha'] = radarobject['GeneralRadarSpec_Object']['CFAR_Angle_alpha']
        
        specifications['PrecodingMatrix'] = mimo_Functions.AD_matrix(NPulse=specifications['NPulse'],
                                                                    M=len(radarobject['TX']),
                                                                    tech=specifications['MIMO_Tech'])
        specifications['DopplerProcessingMIMODemod'] = radarobject['GeneralRadarSpec_Object']['DopplerProcessingMIMODemod']
        
        specifications['ADC_peak2peak'] = radarobject['GeneralRadarSpec_Object']['ADC_peak2peak']
        specifications['ADC_levels'] = radarobject['GeneralRadarSpec_Object']['ADC_levels']
        specifications['ADC_ImpedanceFactor'] = radarobject['GeneralRadarSpec_Object']['ADC_ImpedanceFactor']
        specifications['ADC_LNA_Gain'] = 10**(radarobject['GeneralRadarSpec_Object']['ADC_LNA_Gain_dB']/10)
        specifications['ADC_SaturationEnabled'] = radarobject['GeneralRadarSpec_Object']['ADC_SaturationEnabled']
        specifications['RF_NoiseFiguredB'] = radarobject['GeneralRadarSpec_Object']['RF_NoiseFiguredB']
        specifications['RF_AnalogNoiseFilter_Bandwidth'] = radarobject['GeneralRadarSpec_Object']['RF_AnalogNoiseFilter_Bandwidth_MHz']*1e6
        specifications['MIMO_Antenna_Azimuth_Elevation_Order']=radarobject['GeneralRadarSpec_Object']['antenna2azelIndex']
        specifications['matrix_world']=radarobject['GeneralRadarSpec_Object'].matrix_world.decompose()
        k=0
        global_location_Center = Vector((0,0,0))
        global_location_TX = []
        for itx,txobj in enumerate(radarobject['TX']):
          global_location, global_rotation, global_scale = txobj.matrix_world.decompose()
          global_location_TX.append(global_location)
          global_location_Center += global_location
          k+=1
        global_location_RX = []
        for irx,rxobj in enumerate(radarobject['RX']):
          global_location, global_rotation, global_scale = rxobj.matrix_world.decompose()
          global_location_RX.append(global_location)
          global_location_Center += global_location
          k+=1
        global_location_Center /= k
        specifications['global_location_TX_RX_Center'] = [global_location_TX,global_location_RX,global_location_Center]
        azindex = []
        elindex = []
        for itx,txobj in enumerate(radarobject['TX']):
          # global_location, global_rotation, global_scale = txobj.matrix_world.decompose()
          local_location, local_rotation, local_scale = txobj.matrix_local.decompose()
          local_location_HW = local_location / (ssp.constants.LightSpeed/radarobject['GeneralRadarSpec_Object']["Center_Frequency_GHz"]/1e9)* 2
          azTx=round(local_location_HW.x)
          elTx=round(local_location_HW.y)
          # print("itx,local_location:",itx,local_location,txobj.name)
          for irx,rxobj in enumerate(radarobject['RX']):
              local_location, local_rotation, local_scale = rxobj.matrix_local.decompose()
              local_location_HW = local_location / (ssp.constants.LightSpeed/radarobject['GeneralRadarSpec_Object']["Center_Frequency_GHz"]/1e9) * 2
              azRx=round(local_location_HW.x)
              elRx=round(local_location_HW.y)
              # print(iradar,azTx+azRx,elRx+elTx)
              azindex.append(azTx+azRx)
              elindex.append(elTx+elRx)
        #       print("irx,local_location:",irx,local_location,rxobj.name)
        # print("azindex:",azindex)
        # print("elindex:",elindex)


        azindex = azindex - np.min(azindex)+1
        elindex = elindex - np.min(elindex)+1
        antennaIndex2VAx = np.zeros((len(radarobject['TX']),len(radarobject['RX'])))
        antennaIndex2VAy = np.zeros((len(radarobject['TX']),len(radarobject['RX'])))
        k=0
        for itx in range(antennaIndex2VAx.shape[0]):
          for irx in range(antennaIndex2VAx.shape[1]):
            antennaIndex2VAx[itx,irx] = azindex[k]-1
            antennaIndex2VAy[itx,irx] = elindex[k]-1
            k+=1

        specifications['MIMO_AntennaIndex2VA']=[antennaIndex2VAx,antennaIndex2VAy,np.max(elindex),np.max(azindex)]
        antenna_Pos0_Wavelength_TX=[]
        for itx,txobj in enumerate(radarobject['TX']):
          local_location, local_rotation, local_scale = txobj.matrix_local.decompose()
          local_location_HW = local_location / (ssp.constants.LightSpeed/radarobject['GeneralRadarSpec_Object']["Center_Frequency_GHz"]/1e9)
          antenna_Pos0_Wavelength_TX.append(local_location_HW)
        antenna_Pos0_Wavelength_RX=[]
        for irx,rxobj in enumerate(radarobject['RX']):
              local_location, local_rotation, local_scale = rxobj.matrix_local.decompose()
              local_location_HW = local_location / (ssp.constants.LightSpeed/radarobject['GeneralRadarSpec_Object']["Center_Frequency_GHz"]/1e9)
              antenna_Pos0_Wavelength_RX.append(local_location_HW)
        specifications['antenna_Pos0_Wavelength']=[antenna_Pos0_Wavelength_TX,antenna_Pos0_Wavelength_RX]


        PosIndex = []
        for itx,txobj in enumerate(radarobject['TX']):
          # global_location, global_rotation, global_scale = txobj.matrix_world.decompose()
          local_location, local_rotation, local_scale = txobj.matrix_local.decompose()
          local_location_HW = local_location / (ssp.constants.LightSpeed/radarobject['GeneralRadarSpec_Object']["Center_Frequency_GHz"]/1e9)* 2
          azTx=local_location_HW.x
          elTx=local_location_HW.y
          # print("itx,local_location:",itx,local_location,txobj.name)
          for irx,rxobj in enumerate(radarobject['RX']):
              local_location, local_rotation, local_scale = rxobj.matrix_local.decompose()
              local_location_HW = local_location / (ssp.constants.LightSpeed/radarobject['GeneralRadarSpec_Object']["Center_Frequency_GHz"]/1e9) * 2
              azRx=local_location_HW.x
              elRx=local_location_HW.y
              PosIndex.append([azTx+azRx,elTx+elRx,itx,irx])
        specifications['Local_location_TXplusRX_Center'] = PosIndex

        # x = np.zeros((np.max(elindex),np.max(azindex)))
        # for itx,txobj in enumerate(radarobject['TX']):
        #   for irx,rxobj in enumerate(radarobject['RX']):
        #     x[int(antennaIndex2VAy[itx,irx]),int(antennaIndex2VAx[itx,irx])]=1

        # print(iradar,azindex,elindex,np.max(azindex),np.max(elindex),x)
        # specifications['RangePulseRX']= np.zeros((specifications['N_ADC'],specifications['NPulse'],len(Suite_Position[isuite]['Radar'][iradar]['RX-Position'])),dtype=np.complex128)
        
        specifications['RadarTiming'] = RadarTiming(t_start_radar=radarobject['GeneralRadarSpec_Object']['t_start_radar'], 
                                                    t_start_manual_restart_tx=0,
                                                    t_last_pulse=0.0,
                                                    t_current_pulse=0, 
                                                    pri_sequence=[specifications['PRI']], 
                                                    n_pulse=0,
                                                    n_last_cpi=0)
        specifications['CPI_Buffer']=[]
        specifications['RadarBuffer']=RadarBuffer(specifications['NPulse'])
        specifications['SaveSignalGenerationTime']=radarobject['GeneralRadarSpec_Object']['SaveSignalGenerationTime']
        specifications['MaxRangeScatter']=radarobject['GeneralRadarSpec_Object']['MaxRangeScatter']
        specifications['continuousCPIsTrue_oneCPIpeerFrameFalse']=radarobject['GeneralRadarSpec_Object']['continuousCPIsTrue_oneCPIpeerFrameFalse']

        tx_positions,rx_positions=radarobject['GeneralRadarSpec_Object']["TXRXPos"]
        vainfo = ssp.radar.utils.virtualArray_info(tx_positions,rx_positions)
        specifications['ULA_TXRX_Lx_Ly_NonZ']=vainfo
        radarSpecifications.append(specifications)
      RadarSpecifications.append(radarSpecifications)
      

    
    # if os.path.exists('frames'):
    #     shutil.rmtree('frames')
    # os.makedirs('frames')
    ssp.RadarSpecifications = RadarSpecifications

def open_Blend(file):
  bpy.ops.wm.open_mainfile(filepath=file)
def applyDecimate(obj, decimateFactor):
    if obj.data.shape_keys is not None:
        bpy.context.view_layer.objects.active = obj 
        blocks = obj.data.shape_keys.key_blocks
        for ind in reversed(range(len(blocks))):
            obj.active_shape_key_index = ind
            bpy.ops.object.shape_key_remove()
    if len(obj.data.vertices) * decimateFactor > 10: 
        decimate_mod = obj.modifiers.new(type='DECIMATE', name='decimate')
        decimate_mod.ratio = decimateFactor
        decimate_mod.use_collapse_triangulate = True
        bpy.ops.object.modifier_apply(modifier='decimate')

def cleanAllDecimateModifiers(obj):
    for m in obj.modifiers:
        if(m.type=="DECIMATE"):
            obj.modifiers.remove(modifier=m)

def mesh2triangles(mesh):
    out = []
    mesh.calc_loop_triangles()
    for tri in mesh.loop_triangles: 
        xyz0 = mesh.vertices[tri.vertices[0]].co.to_tuple()
        xyz1 = mesh.vertices[tri.vertices[1]].co.to_tuple()
        xyz2 = mesh.vertices[tri.vertices[2]].co.to_tuple()
        out.append([xyz0,xyz1,xyz2])
    return out
def createMeshesCollection_fromTrianglesRCS0(triangleList,vertex_colors,sigma0=1.0,addMatrial=False):
  bpy.ops.object.empty_add(type='PLAIN_AXES', location=(0, 0, 0))
  p = bpy.context.object
  k=0
  for mesh in triangleList:
      for triangle in mesh:
          if vertex_colors is None:
            color=1
          else:  
            color = vertex_colors[k]
          # if color<128:
          #     color=0
          # else:
          #     color=255
          mesh = bpy.data.meshes.new(f"Rectangle_Mesh_{k}")
          obj = bpy.data.objects.new(f"Rectangle_{k}", mesh)
          bpy.context.collection.objects.link(obj)
          faces = [(0, 1, 2)]
          mesh.from_pydata(triangle, [], faces)
          mesh.update()
          if addMatrial:
              mat = bpy.data.materials.new(name=f"Material_{k}")
              mat.use_nodes = True
              bsdf = mat.node_tree.nodes["Principled BSDF"]
              mat.node_tree.nodes["Principled BSDF"].inputs["Base Color"].default_value = (color, color, color, 1)  
              if obj.data.materials:
                  obj.data.materials[0] = mat
              else:
                  obj.data.materials.append(mat)
          obj["RCS0"] = color*sigma0
          obj.parent = p
          k+=1

def scale_unit_sphere(x,y,z):        
  x=x-np.mean(x)
  x=x/np.max(x)
  y=y-np.mean(y)
  y=y/np.max(y)
  z=z-np.mean(z)
  z=z/np.max(z)
  return x,y,z
def vectorize_triangleList(triangleList):        
  x,y,z=[],[],[]
  for mesh in triangleList:
      for triangle in mesh:
          v0, v1, v2 = triangle
          x.append((v0[0]+v1[0]+v2[0])/3.0)
          y.append((v0[1]+v1[1]+v2[1])/3.0)
          z.append((v0[2]+v1[2]+v2[2])/3.0)
  x=np.array(x)
  y=np.array(y)
  z=np.array(z)
  return x,y,z

def exportBlenderTriangles():
  frame = ssp.config.CurrentFrame
  out = []
  bpy.context.scene.frame_set(frame)
  bpy.context.view_layer.update()
  for obj in bpy.context.scene.objects:
    if obj.type == 'MESH':
      depgraph = bpy.context.evaluated_depsgraph_get()
      bm = bmesh.new()
      bm.verts.ensure_lookup_table()
      bm.from_object(obj, depgraph)
      bm.transform(obj.matrix_world)
      mesh = bpy.data.meshes.new('new_mesh')
      bm.to_mesh(mesh)
      trianglesList=mesh2triangles(mesh)
      out.append(trianglesList)
  return out

def decimate_scene(decimation_ratio = 0.5):
  for obj in bpy.context.selected_objects:
      if obj.type == 'MESH':
          # Add a Decimate modifier
          decimate_modifier = obj.modifiers.new(name="Decimate", type='DECIMATE')
          decimate_modifier.ratio = decimation_ratio
          
          # Apply the Decimate modifier
          bpy.ops.object.modifier_apply(modifier=decimate_modifier.name)
def decimate_scene_all(decimation_ratio = 0.5):
  for obj in bpy.data.objects:
      if obj.type == 'MESH':
          # Set the object as active
          bpy.context.view_layer.objects.active = obj
          obj.select_set(True)

          # Add a decimate modifier
          modifier = obj.modifiers.new(name="Decimate", type='DECIMATE')
          modifier.ratio = decimation_ratio

          # Apply the decimate modifier
          bpy.ops.object.modifier_apply(modifier="Decimate")

          # Deselect the object
          obj.select_set(False)

def plot1D(ax,x,xlabel="",ylabel="",title="",ylim=[]):
  ax.cla()
  ax.plot(x)
  ax.set_xlabel(xlabel)
  ax.set_ylabel(ylabel)
  ax.set_title(title)
  if len(ylim)==2:
    ax.set_ylim(ylim)
  ax.grid(True)
def plot2D(ax,x,xlabel="",ylabel="",title=""):
  ax.cla()
  ax.imshow(x)
  ax.set_xlabel(xlabel)
  ax.set_ylabel(ylabel)
  ax.set_title(title)
  

def open_folder(folder_path):
    if not os.path.exists(folder_path):
        print("Folder does not exist.")
        return

    # Detect the OS and use the appropriate method to open the folder
    current_os = platform.system()

    if current_os == "Windows":
        # For Windows
        os.startfile(folder_path)
    elif current_os == "Linux":
        # For Linux
        subprocess.run(['xdg-open', folder_path])
    elif current_os == "Darwin":  # For macOS
        subprocess.run(['open', folder_path])
    else:
        print("Unsupported operating system.")


def open_temp_folder():
  open_folder(ssp.config.temp_folder)

def readImage_tempfolder(imagefilename,resize=(400, 120),display=False):
  moonTexture_file = os.path.join(ssp.config.temp_folder, imagefilename)
  image = cv2.imread(moonTexture_file)
  if resize is not None:
    image = cv2.resize(image, resize)
  gray_image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
  if display:
    cv2.imshow("Grayscale Image", gray_image)
    cv2.waitKey(0)
    cv2.destroyAllWindows()
  return gray_image
class RadarTiming:
    def __init__(self, t_start_radar: float, t_start_manual_restart_tx: float, t_last_pulse: float,
                 t_current_pulse: float, pri_sequence: list, n_pulse: int, n_last_cpi: int):
        self.t_start_radar = np.float64(t_start_radar)
        self.t_start_manual_restart_tx = np.float64(t_start_manual_restart_tx)
        self.t_last_pulse = np.float64(t_last_pulse)
        self.t_current_pulse = np.float64(t_current_pulse)
        self.pri_sequence = np.array(pri_sequence, dtype=np.float64)
        self.n_pulse = n_pulse
        self.n_last_cpi = np.int64(n_last_cpi)
        self.Frame_PulseTimes = np.array([],dtype=np.float64)
        self.Frame_PulseWidths = np.array([],dtype=np.float64)
        self.Started = False

    def update(self,currentTime,FPS): # [ )
      self.Frame_PulseTimes = np.array([],dtype=np.float64)
      self.Frame_PulseWidths = np.array([],dtype=np.float64)
      T = 1/float(FPS) # 1/24 = 0.04167
      if self.t_start_radar >= currentTime+T:
        return
      if self.Started == False:
        if self.t_start_radar >= currentTime:
          self.Started = True
          self.t_current_pulse = self.t_start_radar
          self.n_pulse = 1
          self.Frame_PulseTimes = np.append(self.Frame_PulseTimes, self.t_current_pulse)
          self.Frame_PulseWidths = np.append(self.Frame_PulseWidths, self.pri_sequence[0])
      if self.Started == False:
        return
        
      while 1:
        k=self.n_pulse-1
        k=0
        CurrentPRI = self.pri_sequence[k]
        if self.t_current_pulse + CurrentPRI >= currentTime+T:
          break
        self.t_current_pulse += CurrentPRI
        self.n_pulse+=1
        self.Frame_PulseTimes = np.append(self.Frame_PulseTimes, self.t_current_pulse)
        self.Frame_PulseWidths = np.append(self.Frame_PulseWidths, self.pri_sequence[0])
        
# class RadarTiming:
#     def __init__(self, t_start_radar: float, pri_sequence: list, cpi_length: int):
#         self.t_start_radar = np.float64(t_start_radar)
#         self.pri_sequence = np.array(pri_sequence, dtype=np.float64)
#         self.cpi_length = cpi_length
#         self.t_current_pulse = t_start_radar
#         self.n_pulse = 0
#         self.Frame_PulseTimes = np.array([], dtype=np.float64)
#         self.Started = False
#         self.pri_index = 0

#     def update(self, currentTime, FPS):
#         self.Frame_PulseTimes = []  # Use list for efficiency during appending
#         T = 1 / float(FPS)  # Frame duration

#         # Check if radar should start
#         if self.t_start_radar >= currentTime + T:
#             return

#         # Initialize radar start if needed
#         if not self.Started:
#             if self.t_start_radar >= currentTime:
#                 self.Started = True
#                 self.t_current_pulse = self.t_start_radar
#                 self.n_pulse = 1
#                 self.Frame_PulseTimes.append(self.t_current_pulse)

#         if not self.Started:
#             return

#         # Process pulses within the current frame
#         while True:
#             current_pri = self.pri_sequence[self.pri_index]
#             if self.t_current_pulse + current_pri >= currentTime + T:
#                 break

#             self.t_current_pulse += current_pri
#             self.n_pulse += 1
#             self.Frame_PulseTimes.append(self.t_current_pulse)

#             # Move to the next PRI, wrapping around if needed
#             self.pri_index = (self.pri_index + 1) % len(self.pri_sequence)

#         # Convert the list to numpy array after appending
#         self.Frame_PulseTimes = np.array(self.Frame_PulseTimes, dtype=np.float64)


class RadarBuffer:
    def __init__(self, cpi_length: int):
        self.cpi_length = cpi_length
        self.buffer = []
        self.remaining_pulses = []

    def buffer_pulses(self, pulses, option=1):
        if option == 1:
            # Buffer all generated pulses
            self.buffer.extend(pulses)
        elif option == 2:
            # Buffer only complete CPIs and store remaining pulses
            total_pulses = self.remaining_pulses + list(pulses)
            while len(total_pulses) >= self.cpi_length:
                complete_cpi = total_pulses[:self.cpi_length]
                self.buffer.append(complete_cpi)
                total_pulses = total_pulses[self.cpi_length:]
            self.remaining_pulses = total_pulses

    def get_buffered_cpis(self):
        return self.buffer

    def get_remaining_pulses(self):
        return self.remaining_pulses


def channels_info(path_d_drate_amp,isuiteRX=0,iradarRX=0,isuiteTX=0,iradarTX=0):
  fig, ax = plt.subplots(2,2)
  for irx  in range(len(path_d_drate_amp[isuiteRX][iradarRX])):
    PathNumber = [len(path_d_drate_amp[isuiteRX][iradarRX][irx][isuiteTX][iradarTX][itx]) for itx in range(len(path_d_drate_amp[isuiteRX][iradarRX][irx][isuiteTX][iradarTX]))]
    ax[0,0].plot(PathNumber)
    ax[0,0].set_xlabel('TX index')
    ax[0,0].set_ylabel('Path Number from RXi (legend) ') 
  itx,irx=0,0
  Channel_d_fd_amp=np.array([[float(d),float(dr),float(a)] for d,dr,a,m in path_d_drate_amp[isuiteRX][iradarRX][irx][isuiteTX][iradarTX][itx]])
  if len(Channel_d_fd_amp)==0:
    plt.show()
    return
  ax[0,1].plot(Channel_d_fd_amp[:,0]/(3e8/1e9),20*np.log10(Channel_d_fd_amp[:,2]),'.')
  ax[0,1].set_xlabel('delay (nsec)')
  ax[0,1].set_ylabel('amp (dB) Tx0RX0') 
  ax[1,1].plot(Channel_d_fd_amp[:,0]/(3e8/1e9),Channel_d_fd_amp[:,1]/0.005,'.')
  ax[1,1].set_xlabel('delay (nsec)')
  ax[1,1].set_ylabel('Doppler (Hz) for 5mm wavelength Tx0RX0') 
  d,dr,a,m = path_d_drate_amp[isuiteRX][iradarRX][irx][isuiteTX][iradarTX][itx][0]
  all_phase = []
  for irx  in range(len(path_d_drate_amp[isuiteRX][iradarRX])):
    for itx in range(len(path_d_drate_amp[isuiteRX][iradarRX][irx][isuiteTX][iradarTX])):
      for di,dri,ai,mi in path_d_drate_amp[isuiteRX][iradarRX][irx][isuiteTX][iradarTX][itx]:
        if mi==m:
          all_phase.append((di-d)/0.005*360)
          break
  ax[1,0].plot(all_phase,'.')
  ax[1,0].plot(sorted(all_phase),'.')
  ax[1,0].set_xlabel('TXRX pair index')
  ax[1,0].set_ylabel('Phase (deg) for 5mm wavelength for scatter 0') 
  plt.show()
  
#   Channel_d_fd_amp=np.array([[float(d),float(dr),float(a)] for d,dr,a,m in path_d_drate_amp[isuiteRX][iradarRX][irx][isuiteTX][iradarTX][itx]])
#   fig, ax = plt.subplots(1,1)
#   ax = fig.add_subplot(1, 1, 1 , projection='3d')
#   ax.scatter(Channel_d_fd_amp[:,0],Channel_d_fd_amp[:,1],Channel_d_fd_amp[:,2])
#   ax.set_xlabel('Distance (m)')
#   ax.set_ylabel('Distance Rate (m/s)') 
#   ax.set_zlabel('Amplitude (v)')
#   # ax.set_box_aspect([1,1,1])  
#   plt.show()
#   # plt.draw()
#   # plt.pause(.1)

def set_axes_equal(ax):
    """Set 3D plot axes to equal scale so the units are equal in all directions."""
    x_limits = ax.get_xlim3d()
    y_limits = ax.get_ylim3d()
    z_limits = ax.get_zlim3d()

    x_range = abs(x_limits[1] - x_limits[0])
    y_range = abs(y_limits[1] - y_limits[0])
    z_range = abs(z_limits[1] - z_limits[0])

    max_range = max(x_range, y_range, z_range)

    x_middle = np.mean(x_limits)
    y_middle = np.mean(y_limits)
    z_middle = np.mean(z_limits)

    ax.set_xlim3d([x_middle - max_range / 2, x_middle + max_range / 2])
    ax.set_ylim3d([y_middle - max_range / 2, y_middle + max_range / 2])
    ax.set_zlim3d([z_middle - max_range / 2, z_middle + max_range / 2])